import pytest

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.kernel_ridge import primal_weights_weighted_kernel_ridge
from himalaya.kernel_ridge import predict_weighted_kernel_ridge


def _create_dataset(backend):
    n_samples, n_targets = 30, 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in [100, 200]
    ]
    Ks = backend.stack([backend.matmul(X, X.T) for X in Xs])
    Y = backend.asarray(backend.randn(n_samples, n_targets), backend.float64)
    dual_weights = backend.asarray(backend.randn(n_samples, n_targets),
                                   backend.float64)
    exp_deltas = backend.asarray(backend.rand(Ks.shape[0], n_targets),
                                 backend.float64)
    deltas = backend.log(exp_deltas)

    return Xs, Ks, Y, deltas, dual_weights


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_predict_weighted_kernel_ridge(backend):
    backend = set_backend(backend)
    Xs, Ks, _, deltas, dual_weights = _create_dataset(backend)

    primal_weights = primal_weights_weighted_kernel_ridge(
        dual_weights, deltas, Xs)
    predictions_primal = backend.stack(
        [X @ backend.asarray(w) for X, w in zip(Xs, primal_weights)]).sum(0)

    predictions_dual = predict_weighted_kernel_ridge(Ks, dual_weights, deltas)

    assert_array_almost_equal(predictions_primal, predictions_dual)



import pytest
import sklearn.utils.estimator_checks

from himalaya.backend import set_backend
from himalaya.backend import get_backend
from himalaya.backend import ALL_BACKENDS

from himalaya.lasso import SparseGroupLassoCV

###############################################################################
# scikit-learn.utils.estimator_checks


class SparseGroupLassoCV_(SparseGroupLassoCV):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, groups=None, l1_regs=(0, 0.1), l21_regs=(0, 0.1),
                 solver="proximal_gradient", solver_params=None, cv=2):
        super().__init__(groups=groups, l1_regs=l1_regs, l21_regs=l21_regs,
                         solver=solver, solver_params=solver_params, cv=cv)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        from himalaya.validation import check_array
        from himalaya.scoring import r2_score
        backend = get_backend()

        y_pred = super().predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return backend.to_numpy(
                r2_score(y_true[:, None], y_pred[:, None])[0])
        else:
            return backend.to_numpy(r2_score(y_true, y_pred))


@sklearn.utils.estimator_checks.parametrize_with_checks([
    SparseGroupLassoCV_(),
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_check_estimator(estimator, check, backend):
    backend = set_backend(backend)
    check(estimator)



from abc import ABC, abstractmethod

from sklearn.base import BaseEstimator, RegressorMixin, MultiOutputMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.model_selection import check_cv

from ._solvers import RIDGE_SOLVERS
from ._random_search import GROUP_RIDGE_SOLVERS
from ._random_search import solve_ridge_cv_svd

from ..validation import check_array
from ..validation import _get_string_dtype
from ..backend import get_backend
from ..scoring import r2_score


class _BaseRidge(ABC, MultiOutputMixin, RegressorMixin, BaseEstimator):
    """Base class for ridge estimators"""

    @property
    @classmethod
    @abstractmethod
    def ALL_SOLVERS(cls):
        ...

    def _call_solver(self, **direct_params):
        if self.solver not in self.ALL_SOLVERS:
            raise ValueError("Unknown solver=%r." % self.solver)

        function = self.ALL_SOLVERS[self.solver]
        solver_params = self.solver_params or {}

        # check duplicated parameters
        intersection = set(direct_params.keys()).intersection(
            set(solver_params.keys()))
        if intersection:
            raise ValueError(
                'Parameters %s should not be given in solver_params, since '
                'they are either fixed or have a direct parameter in %s.' %
                (intersection, self.__class__.__name__))

        return function(**direct_params, **solver_params)

    def _more_tags(self):
        return {'requires_y': True}


class Ridge(_BaseRidge):
    """Ridge regression.

    Solve the ridge regression::

        b* = argmin_b ||X @ b - Y||^2 + alpha ||b||^2.

    Parameters
    ----------
    alpha : float, or array of shape (n_targets, )
        L2 regularization parameter.

    fit_intercept : boolean
        Whether to fit an intercept.
        If False, X and Y must be zero-mean over samples.

    solver : str
        Algorithm used during the fit, in {"svd"}.

    solver_params : dict or None
        Additional parameters for the solver.
        See more details in the docstring of the function:
        ``Ridge.ALL_SOLVERS[solver]``

    Attributes
    ----------
    coef_ : array of shape (n_features) or (n_features, n_targets)
        Ridge coefficients.

    intercept_ : float or array of shape (n_targets, )
        Intercept. Only present if fit_intercept is True.

    n_features_in_ : int
        Number of features used during the fit.

    dtype_ : str
        Dtype of input data.

    Examples
    --------
    >>> from himalaya.ridge import Ridge
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 3
    >>> X = np.random.randn(n_samples, n_features)
    >>> Y = np.random.randn(n_samples, n_targets)
    >>> model = Ridge()
    >>> model.fit(X, Y)
    Ridge()
    """
    ALL_SOLVERS = RIDGE_SOLVERS

    def __init__(self, alpha=1, fit_intercept=False, solver="svd",
                 solver_params=None):
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.solver = solver
        self.solver_params = solver_params

    def fit(self, X, y=None):
        """Fit the model.

        Parameters
        ----------
        X : array of shape (n_samples, n_features)
            Training data.

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        # backend = get_backend()
        X = check_array(X, ndim=2)
        self.dtype_ = _get_string_dtype(X)
        y = check_array(y, dtype=self.dtype_, ndim=[1, 2])
        if X.shape[0] != y.shape[0]:
            raise ValueError("Inconsistent number of samples.")

        self.n_features_in_ = X.shape[1]

        ravel = False
        if y.ndim == 1:
            y = y[:, None]
            ravel = True

        # ------------------ call the solver
        tmp = self._call_solver(X=X, Y=y, alpha=self.alpha,
                                fit_intercept=self.fit_intercept)
        if self.fit_intercept:
            self.coef_, self.intercept_ = tmp
        else:
            self.coef_ = tmp

        if ravel:
            self.coef_ = self.coef_[:, 0]
            if self.fit_intercept:
                self.intercept_ = self.intercept_[0]

        return self

    def predict(self, X):
        """Predict using the model.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features)
            Testing features.

        Returns
        -------
        Y_hat : array of shape (n_samples,) or (n_samples, n_targets)
            Returns predicted values.
        """
        check_is_fitted(self)
        backend = get_backend()
        X = check_array(X, dtype=self.dtype_, ndim=2)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                'Different number of features in X than during fit.')

        Y_hat = backend.to_cpu(X) @ backend.to_cpu(self.coef_)
        if self.fit_intercept:
            Y_hat += backend.to_cpu(self.intercept_)
        return Y_hat

    def score(self, X, y):
        """Return the coefficient of determination R^2 of the prediction.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features)
            Testing features.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            True values for X.

        Returns
        -------
        score : array of shape (n_targets, )
            R^2 of self.predict(X) versus y.
        """
        y_pred = self.predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return r2_score(y_true[:, None], y_pred[:, None])[0]
        else:
            return r2_score(y_true, y_pred)


class RidgeCV(Ridge):
    """Ridge regression with efficient cross-validation over alpha.

    Solve the ridge regression::

        b* = argmin_b ||X @ b - Y||^2 + alpha ||b||^2,

    with a grid-search over cross-validation to find the best alpha.

    Parameters
    ----------
    alphas : array of shape (n_alphas, )
        List of L2 regularization parameter to try.

    fit_intercept : boolean
        Whether to fit an intercept.
        If False, X and Y must be zero-mean over samples.

    solver : str
        Algorithm used during the fit, "svd" only for now.

    solver_params : dict or None
        Additional parameters for the solver.
        See more details in the docstring of the function:
        ``RidgeCV.ALL_SOLVERS[solver]``

    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.

    Y_in_cpu : bool
        If True, keep the target values ``y`` in CPU memory (slower).

    Attributes
    ----------
    coef_ : array of shape (n_features) or (n_features, n_targets)
        Ridge coefficients.

    intercept_ : float or array of shape (n_targets, )
        Intercept. Only returned when fit_intercept is True.

    best_alphas_ : array of shape (n_targets, )
        Selected best hyperparameter alphas.

    cv_scores_ : array of shape (n_targets, )
        Cross-validation scores averaged over splits, for the best alpha.

    n_features_in_ : int
        Number of features used during the fit.

    Examples
    --------
    >>> from himalaya.ridge import RidgeCV
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 3
    >>> X = np.random.randn(n_samples, n_features)
    >>> Y = np.random.randn(n_samples, n_targets)
    >>> clf = RidgeCV()
    >>> clf.fit(X, Y)
    RidgeCV()
    """
    ALL_SOLVERS = dict(svd=solve_ridge_cv_svd)

    def __init__(self, alphas=[0.1, 1], fit_intercept=False, solver="svd",
                 solver_params=None, cv=5, Y_in_cpu=False):
        self.alphas = alphas
        self.fit_intercept = fit_intercept
        self.solver = solver
        self.solver_params = solver_params
        self.cv = cv
        self.Y_in_cpu = Y_in_cpu

    def fit(self, X, y=None):
        """Fit ridge regression model

        Parameters
        ----------
        X : array of shape (n_samples, n_features).
            Training data.

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        # backend = get_backend()
        X = check_array(X, ndim=2)
        self.dtype_ = _get_string_dtype(X)
        device = "cpu" if self.Y_in_cpu else None
        y = check_array(y, dtype=self.dtype_, ndim=[1, 2], device=device)
        if X.shape[0] != y.shape[0]:
            raise ValueError("Inconsistent number of samples.")

        alphas = check_array(self.alphas, dtype=self.dtype_, ndim=1)
        self.n_features_in_ = X.shape[1]

        ravel = False
        if y.ndim == 1:
            y = y[:, None]
            ravel = True

        cv = check_cv(self.cv)

        # ------------------ call the solver
        tmp = self._call_solver(X=X, Y=y, cv=cv, alphas=alphas,
                                fit_intercept=self.fit_intercept,
                                Y_in_cpu=self.Y_in_cpu)
        if self.fit_intercept:
            self.best_alphas_, self.coef_, self.cv_scores_ = tmp[:3]
            self.intercept_, = tmp[3:]
        else:
            self.best_alphas_, self.coef_, self.cv_scores_ = tmp

        self.cv_scores_ = self.cv_scores_[0]

        if ravel:
            self.coef_ = self.coef_[:, 0]
            if self.fit_intercept:
                self.intercept_ = self.intercept_[0]

        return self


###############################################################################
###############################################################################
###############################################################################
###############################################################################


class GroupRidgeCV(_BaseRidge):
    """Group ridge regression with cross-validation.

    Solve the group-regularized ridge regression::

        b* = argmin_b ||Z @ b - Y||^2 + ||b||^2

    where the feature space X_i is scaled by a group scaling ::

        Z_i = exp(deltas[i] / 2) X_i

    The solver optimizes the log group scalings ``deltas`` over
    cross-validation, using random search (``solver="random_search"``).

    Parameters
    ----------
    groups : array of shape (n_features, ), "input", or None
        Encoding of the group of each feature. If None, all features are
        gathered in one group, and the problem is equivalent to RidgeCV.
        If "input", the input features ``X`` should be a list of 2D arrays,
        corresponding to each group.

    solver : str
        Algorithm used during the fit, only "random_search" for now.

    solver_params : dict or None
        Additional parameters for the solver.
        See more details in the docstring of the function:
        ``GroupRidgeCV.ALL_SOLVERS[solver]``

    fit_intercept : boolean
        Whether to fit an intercept.
        If False, X and Y must be zero-mean over samples.

    cv : int or scikit-learn splitter
        Cross-validation splitter. If an int, KFold is used.

    random_state : int, or None
        Random generator seed. Use an int for deterministic search.

    Y_in_cpu : bool
        If True, keep the target values ``y`` in CPU memory (slower).

    Attributes
    ----------
    coef_ : array of shape (n_features) or (n_features, n_targets)
        Ridge coefficients.

    intercept_ : float or array of shape (n_targets, )
        Intercept. Only returned when fit_intercept is True.

    deltas_ : array of shape (n_groups, n_targets)
        Log of the group scalings.

    cv_scores_ : array of shape (n_iter, n_targets)
        Cross-validation scores, averaged over splits.

    n_features_in_ : int
        Number of features used during the fit.

    dtype_ : str
        Dtype of input data.

    best_alphas_ : array of shape (n_targets, )
        Equal to ``1. / exp(self.deltas_).sum(0)``. For the "random_search"
        solver, it corresponds to the best hyperparameter alphas, assuming that
        each squared group scaling vector sums to one (in particular, it is the
        case when ``solver_params['n_iter']`` is an integer).

    Examples
    --------
    >>> from himalaya.ridge import GroupRidgeCV
    >>> from himalaya.ridge import ColumnTransformerNoStack
    >>> from sklearn.pipeline import make_pipeline

    >>> # create a dataset
    >>> import numpy as np
    >>> n_samples, n_features, n_targets = 10, 5, 3
    >>> X = np.random.randn(n_samples, n_features)
    >>> Y = np.random.randn(n_samples, n_targets)

    >>> # Separate the first three columns and the last two
    >>> # columns, creating two groups of shape (n_samples, n_feature_i).
    >>> from sklearn.preprocessing import StandardScaler
    >>> ct = ColumnTransformerNoStack(
    ...     [("group_1", StandardScaler(), [0, 1, 2]),
    ...      ("group_2", StandardScaler(), slice(3, 5))])

    >>> # A model with automatic groups, as output by ColumnTransformerNoStack
    >>> model = GroupRidgeCV(groups="input")
    >>> pipe = make_pipeline(ct, model)
    >>> _ = pipe.fit(X, Y)
    """
    ALL_SOLVERS = GROUP_RIDGE_SOLVERS

    def __init__(self, groups=None, solver="random_search", solver_params=None,
                 fit_intercept=False, cv=5, random_state=None, Y_in_cpu=False):

        self.groups = groups
        self.solver = solver
        self.solver_params = solver_params
        self.fit_intercept = fit_intercept
        self.cv = cv
        self.random_state = random_state
        self.Y_in_cpu = Y_in_cpu

    def fit(self, X, y=None):
        """Fit the model.

        Parameters
        ----------
        X : array of shape (n_samples, n_features), or list of length \
                (n_groups) with arrays of shape (n_samples, n_features)
            Training data.
            Must be a 2D array if ``groups`` is given.
            Must be a list of 2D arrays if ``groups="input"``.

        y : array of shape (n_samples,) or (n_samples, n_targets)
            Target values.

        Returns
        -------
        self : returns an instance of self.
        """
        backend = get_backend()

        Xs = self._split_groups(X, check=True)
        del X

        self.n_features_in_ = sum(Xi.shape[1] for Xi in Xs)

        self.dtype_ = _get_string_dtype(Xs[0])
        device = "cpu" if self.Y_in_cpu else None
        y = check_array(y, dtype=self.dtype_, ndim=[1, 2], device=device)

        if any([Xi.shape[0] != y.shape[0] for Xi in Xs]):
            raise ValueError("Inconsistent number of samples.")

        ravel = False
        if y.ndim == 1:
            y = y[:, None]
            ravel = True

        cv = check_cv(self.cv)

        # ------------------ call the solver
        tmp = self._call_solver(Xs=Xs, Y=y, cv=cv, return_weights=True,
                                random_state=self.random_state,
                                fit_intercept=self.fit_intercept,
                                Y_in_cpu=self.Y_in_cpu)
        if self.fit_intercept:
            self.deltas_, self.coef_, self.cv_scores_ = tmp[:3]
            self.intercept_, = tmp[3:]
        else:
            self.deltas_, self.coef_, self.cv_scores_ = tmp

        if self.solver == "random_search":
            self.best_alphas_ = 1. / backend.exp(self.deltas_).sum(0)
        else:
            self.best_alphas_ = None

        if ravel:
            self.coef_ = self.coef_[:, 0]
            self.deltas_ = self.deltas_[:, 0]
            if self.fit_intercept:
                self.intercept_ = self.intercept_[0]

        return self

    def predict(self, X, split=False):
        """Predict using the model.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features), or list of length \
                (n_groups) with arrays of shape (n_samples_test, n_features)
            Training data.
            Must be a 2D array if ``groups`` is given.
            Must be a list of 2D arrays if ``groups="input"``.

        split : bool
            If True, the prediction is split on each feature space, and this
            method returns an array with one extra dimension (first dimension).
            The sum over this extra dimension corresponds to split=False.

        Returns
        -------
        Y_hat : array of shape (n_samples,) or (n_samples, n_targets)
            Returns predicted values.
            If parameter split is True, the array is of shape
            (n_groups, n_samples,) or (n_groups, n_samples, n_targets).
        """
        backend = get_backend()
        check_is_fitted(self)

        Xs = self._split_groups(X, dtype=self.dtype_, check=True)

        n_features = sum(Xi.shape[1] for Xi in Xs)
        if n_features != self.n_features_in_:
            raise ValueError(
                'Different number of features in X than during fit.')

        X = backend.to_cpu(backend.concatenate(Xs, 1))
        del Xs

        Y_hat = X @ backend.to_cpu(self.coef_)
        if self.fit_intercept:
            Y_hat += backend.to_cpu(self.intercept_)
        return Y_hat

    def score(self, X, y):
        """Return the coefficient of determination R^2 of the prediction.

        Parameters
        ----------
        X : array of shape (n_samples_test, n_features), or list of length \
                (n_groups) with arrays of shape (n_samples_test, n_features)
            Training data.
            Must be a 2D array if ``groups`` is given.
            Must be a list of 2D arrays if ``groups="input"``.

        y : array-like of shape (n_samples,) or (n_samples, n_targets)
            True values for X.

        Returns
        -------
        score : array of shape (n_targets, )
            R^2 of self.predict(X) versus y.
        """
        y_pred = self.predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return r2_score(y_true[:, None], y_pred[:, None])[0]
        else:
            return r2_score(y_true, y_pred)

    def _split_groups(self, X, check=True, **check_kwargs):
        backend = get_backend()

        # groups defined in X
        if self.groups == "input":
            if check:
                X = [check_array(Xi, ndim=2, **check_kwargs) for Xi in X]
            return X

        # groups defined in self.groups
        if check:
            X = check_array(X, ndim=2, **check_kwargs)
        if self.groups is None:
            groups = backend.zeros((X.shape[1]))
        else:
            groups = self.groups
        groups = backend.asarray(groups)[:]
        Xs = [X[:, groups == u] for u in backend.unique(groups) if u >= 0]
        return Xs



"""
Refine multiple-kernel ridge results
====================================
This example demonstrates how to solve multiple-kernel ridge regression with
hyperparameter random search, then refine the results with hyperparameter
gradient descent.
"""
import numpy as np

from himalaya.backend import set_backend
from himalaya.kernel_ridge import MultipleKernelRidgeCV
from himalaya.kernel_ridge import Kernelizer
from himalaya.kernel_ridge import ColumnKernelizer
from himalaya.kernel_ridge import generate_dirichlet_samples

from sklearn.pipeline import make_pipeline

###############################################################################
# In this example, we use the ``cupy`` backend (GPU).

backend = set_backend("cupy")

###############################################################################
# We can display the ``scikit-learn`` pipeline with an HTML diagram.
from sklearn import set_config
set_config(display='diagram')  # requires scikit-learn 0.23

###############################################################################
# Generate a random dataset
# -------------------------
# - Xs_train : list of arrays of shape (n_samples_train, n_features)
# - Xs_test : list of arrays of shape (n_samples_test, n_features)
# - Y_train : array of shape (n_samples_train, n_targets)
# - Y_test : array of shape (n_repeat, n_samples_test, n_targets)

n_kernels = 4
n_targets = 500

# We create a few kernel weights
rng = np.random.RandomState(42)
kernel_weights_true = generate_dirichlet_samples(n_targets, n_kernels,
                                                 concentration=[.3],
                                                 random_state=rng)
kernel_weights_true = backend.to_numpy(kernel_weights_true)

# Then, we generate a random dataset, using the arbitrary scalings.
n_samples_train = 1000
n_samples_test = 400
n_features_list = np.full(n_kernels, fill_value=1000)

Xs_train, Xs_test = [], []
Y_train, Y_test = None, None
for ii in range(n_kernels):
    n_features = n_features_list[ii]

    X_train = rng.randn(n_samples_train, n_features)
    X_test = rng.randn(n_samples_test, n_features)
    X_train -= X_train.mean(0)
    Xs_train.append(X_train)
    Xs_test.append(X_test)

    weights = rng.randn(n_features, n_targets) / n_features
    weights *= kernel_weights_true[:, ii] ** 0.5

    if ii == 0:
        Y_train = X_train @ weights
        Y_test = X_test @ weights
    else:
        Y_train += X_train @ weights
        Y_test += X_test @ weights

std = Y_train.std(0)[None]
Y_train /= std
Y_test /= std

noise = 0.1
Y_train += rng.randn(n_samples_train, n_targets) * noise
Y_test += rng.randn(n_samples_test, n_targets) * noise
Y_test -= Y_test.mean(0)
Y_train -= Y_train.mean(0)

# Concatenate the feature spaces.
X_train = np.asarray(np.concatenate(Xs_train, 1), dtype="float32")
X_test = np.asarray(np.concatenate(Xs_test, 1), dtype="float32")

###############################################################################
# Prepare the pipeline
# --------------------

# Find the start and end of each feature space X in Xs
start_and_end = np.concatenate([[0], np.cumsum(n_features_list)])
slices = [
    slice(start, end)
    for start, end in zip(start_and_end[:-1], start_and_end[1:])
]

# Create a different ``Kernelizer`` for each feature space.
kernelizers = [("space %d" % ii, Kernelizer(), slice_)
               for ii, slice_ in enumerate(slices)]
column_kernelizer = ColumnKernelizer(kernelizers)

###############################################################################
# Define the random-search model
# ------------------------------
# We use very few iteration on purpose, to make the random search suboptimal,
# and refine it with hyperparameter gradient descent.

solver_params = dict(
    n_iter=5,
    alphas=np.logspace(-10, 10, 41),
)

model_1 = MultipleKernelRidgeCV(kernels="precomputed", solver="random_search",
                                solver_params=solver_params, random_state=rng)
pipe_1 = make_pipeline(column_kernelizer, model_1)

# Fit the model on all targets
pipe_1.fit(X_train, Y_train)

###############################################################################
# Define the gradient-descent model
# ---------------------------------

solver_params = dict(
    max_iter=100,
    hyper_gradient_method="direct",
    max_iter_inner_hyper=10,
    initial_deltas="here_will_go_the_previous_deltas"
)

model_2 = MultipleKernelRidgeCV(kernels="precomputed", solver="hyper_gradient",
                                solver_params=solver_params)
pipe_2 = make_pipeline(column_kernelizer, model_2)

###############################################################################
# Use the random-search to initialize the gradient-descent
# --------------------------------------------------------

# We might want to refine only the best predicting targets, since the
# hyperparameter gradient descent is less efficient over many targets.
top = 60  # top 60%
best_cv_scores = backend.to_numpy(pipe_1[-1].cv_scores_.max(0))
mask = best_cv_scores > np.percentile(best_cv_scores, 100 - top)

pipe_2[-1].solver_params['initial_deltas'] = pipe_1[-1].deltas_[:, mask]
pipe_2.fit(X_train, Y_train[:, mask])

###############################################################################
# Compute predictions on a test set
# ---------------------------------
import matplotlib.pyplot as plt

# use the first model for all targets
test_scores_1 = pipe_1.score(X_test, Y_test)

# use the second model for the refined targets
test_scores_2 = backend.copy(test_scores_1)
test_scores_2[mask] = pipe_2.score(X_test, Y_test[:, mask])

test_scores_1 = backend.to_numpy(test_scores_1)
test_scores_2 = backend.to_numpy(test_scores_2)
plt.figure(figsize=(4, 4))
plt.scatter(test_scores_1, test_scores_2, alpha=0.3)
plt.plot(plt.xlim(), plt.xlim(), color='k', lw=1)
plt.xlabel(r"Base model")
plt.ylabel(r"Refined model")
plt.title("$R^2$ generalization score")
plt.tight_layout()
plt.show()



from himalaya.scoring import r2_score
import pytest
import sklearn.kernel_ridge
import sklearn.utils.estimator_checks

from himalaya.backend import set_backend
from himalaya.backend import get_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.ridge import Ridge
from himalaya.ridge import RidgeCV
from himalaya.ridge import GroupRidgeCV
from himalaya.ridge import solve_group_ridge_random_search


def _create_dataset(backend):
    n_samples, n_features, n_targets = 30, 10, 3
    X = backend.asarray(backend.randn(n_samples, n_features), backend.float64)
    w = backend.asarray(backend.randn(n_features, n_targets), backend.float64)
    Y = X @ w
    Y += backend.asarray(backend.randn(*Y.shape), backend.float64)
    return X, Y


@pytest.mark.parametrize('multitarget', [True, False])
@pytest.mark.parametrize('fit_intercept', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_ridge_vs_scikit_learn(backend, multitarget, fit_intercept):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    if not multitarget:
        Y = Y[:, 0]
    if fit_intercept:
        Y += 10
        X += 10

    for alpha in backend.asarray_like(backend.logspace(0, 3, 7), Y):
        model = Ridge(alpha=alpha, fit_intercept=fit_intercept)
        model.fit(X, Y)

        reference = sklearn.linear_model.Ridge(alpha=backend.to_numpy(alpha),
                                               fit_intercept=fit_intercept)
        reference.fit(backend.to_numpy(X), backend.to_numpy(Y))

        if multitarget:
            assert model.coef_.shape == (X.shape[1], Y.shape[1])
        else:
            assert model.coef_.shape == (X.shape[1], )

        assert_array_almost_equal(model.coef_, reference.coef_.T)
        if fit_intercept:
            assert_array_almost_equal(model.intercept_, reference.intercept_,
                                      decimal=5)
        assert_array_almost_equal(model.predict(X),
                                  reference.predict(backend.to_numpy(X)),
                                  decimal=5)
        assert_array_almost_equal(
            model.score(X, Y).mean(),
            reference.score(backend.to_numpy(X), backend.to_numpy(Y)))


@pytest.mark.parametrize('fit_intercept', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_ridge_cv_vs_scikit_learn(backend, fit_intercept):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)
    y = Y[:, 0]
    del Y
    if fit_intercept:
        y += 10
        X += 1

    alphas = backend.asarray_like(backend.logspace(-2, 3, 21), y)

    model = RidgeCV(alphas=alphas, cv=5, fit_intercept=fit_intercept,
                    solver_params=dict(score_func=r2_score))
    model.fit(X, y)

    reference = sklearn.linear_model.RidgeCV(alphas=backend.to_numpy(alphas),
                                             fit_intercept=fit_intercept, cv=5)
    reference.fit(backend.to_numpy(X), backend.to_numpy(y))
    assert model.coef_.shape == (X.shape[1], )

    assert_array_almost_equal(model.best_alphas_[0], reference.alpha_,
                              decimal=5)
    assert_array_almost_equal(model.coef_, reference.coef_.T)
    assert_array_almost_equal(model.predict(X),
                              reference.predict(backend.to_numpy(X)))
    assert_array_almost_equal(
        model.score(X, y).mean(),
        reference.score(backend.to_numpy(X), backend.to_numpy(y)))


@pytest.mark.parametrize('fit_intercept', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_banded_ridge_cv_vs_ridge_cv(backend, fit_intercept):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-2, 3, 21), Y)
    if fit_intercept:
        Y += 10

    ref = RidgeCV(alphas=alphas, cv=5, fit_intercept=fit_intercept)
    ref.fit(X, Y)

    model = GroupRidgeCV(groups=None, solver_params=dict(alphas=alphas), cv=5,
                          fit_intercept=fit_intercept)
    model.fit(X, Y)

    assert_array_almost_equal(model.best_alphas_, ref.best_alphas_)
    assert_array_almost_equal(model.coef_, ref.coef_)
    assert_array_almost_equal(model.predict(X), ref.predict(X))
    assert_array_almost_equal(model.score(X, Y), ref.score(X, Y))


###############################################################################
###############################################################################
###############################################################################
# scikit-learn.utils.estimator_checks


class Ridge_(Ridge):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        from himalaya.validation import check_array
        from himalaya.scoring import r2_score
        backend = get_backend()

        y_pred = super().predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return backend.to_numpy(
                r2_score(y_true[:, None], y_pred[:, None])[0])
        else:
            return backend.to_numpy(r2_score(y_true, y_pred))


class RidgeCV_(RidgeCV):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, alphas=(0.1, 1), solver="svd", solver_params=None,
                 cv=2):
        super().__init__(alphas=alphas, solver=solver,
                         solver_params=solver_params, cv=cv)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        from himalaya.validation import check_array
        from himalaya.scoring import r2_score
        backend = get_backend()

        y_pred = super().predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.coef_.ndim)

        if y_true.ndim == 1:
            return backend.to_numpy(
                r2_score(y_true[:, None], y_pred[:, None])[0])
        else:
            return backend.to_numpy(r2_score(y_true, y_pred))


# Dirty monkey-patch of n_iter,
# since check_estimator does not allow dict parameters
new_defaults = list(solve_group_ridge_random_search.__defaults__)
new_defaults[0] = 1
solve_group_ridge_random_search.__defaults__ = tuple(new_defaults)


class GroupRidgeCV_(GroupRidgeCV):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, groups=None, solver="random_search", solver_params=None,
                 cv=2, random_state=0):
        super().__init__(groups=groups, solver=solver,
                         solver_params=solver_params, cv=cv,
                         random_state=random_state)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        backend = get_backend()
        return backend.to_numpy(super().score(X, y))


@sklearn.utils.estimator_checks.parametrize_with_checks([
    Ridge_(),
    RidgeCV_(),
    GroupRidgeCV_(),
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_check_estimator(estimator, check, backend):
    backend = set_backend(backend)
    check(estimator)



from functools import partial

try:
    import torch
except ImportError as error:
    import sys
    if "pytest" in sys.modules:  # if run through pytest
        import pytest
        pytest.skip("PyTorch not installed.")
    raise ImportError("PyTorch not installed.") from error

from ._utils import _dtype_to_str

###############################################################################


def apply_argmax(array, argmax, axis):
    """Apply precomputed argmax indices in multi dimension arrays

    array[np.argmax(array)] works fine in dimension 1, but not in higher ones.
    This function tackles this issue.

    Examples
    --------
    >>> import torch
    >>> array = torch.randn(10, 4, 8)
    >>> argmax = torch.argmax(array, axis=1)
    >>> max_ = apply_argmax(array, argmax, axis=1)
    >>> assert torch.all(max_ == torch.max(array, axis=1).values)
    """
    argmax = argmax.unsqueeze(dim=axis)
    max_ = torch.gather(array, dim=axis, index=argmax)
    return torch.squeeze(max_, dim=axis)


def std_float64(X, axis=None, demean=True, keepdims=False):
    """Compute the standard deviation of X with double precision,
    and cast back the result to original dtype.
    """
    X_64 = torch.as_tensor(X, dtype=torch.float64)
    X_std = (X_64 ** 2).sum(dim=axis, dtype=torch.float64)
    if demean:
        X_std -= X_64.sum(axis, dtype=torch.float64) ** 2 / X.shape[axis]
    X_std = X_std ** .5
    X_std /= (X.shape[axis] ** .5)

    X_std = torch.as_tensor(X_std, dtype=X.dtype, device=X.device)
    if keepdims:
        X_std = X_std.unsqueeze(dim=axis)

    return X_std


def mean_float64(X, axis=None, keepdims=False):
    """Compute the mean of X with double precision,
    and cast back the result to original dtype.
    """
    X_mean = X.sum(axis, dtype=torch.float64) / X.shape[axis]

    X_mean = torch.as_tensor(X_mean, dtype=X.dtype, device=X.device)
    if keepdims:
        X_mean = X_mean.unsqueeze(dim=axis)
    return X_mean


###############################################################################

name = "torch"
argmax = torch.argmax
randn = torch.randn
rand = torch.rand
matmul = torch.matmul
stack = torch.stack
abs = torch.abs
sum = torch.sum
sqrt = torch.sqrt
any = torch.any
all = torch.all
nan = torch.tensor(float('nan'))
inf = torch.tensor(float('inf'))
isnan = torch.isnan
isinf = torch.isinf
logspace = torch.logspace
concatenate = torch.cat
bool = torch.bool
int32 = torch.int32
float32 = torch.float32
float64 = torch.float64
log = torch.log
exp = torch.exp
arange = torch.arange
unique = torch.unique
einsum = torch.einsum
tanh = torch.tanh
power = torch.pow
prod = torch.prod
sign = torch.sign
clip = torch.clamp
finfo = torch.finfo


def atleast_1d(array):
    array = asarray(array)
    if array.ndim == 0:
        array = array[None]
    return array


def flip(array, axis=0):
    return torch.flip(array, dims=[axis])


def sort(array):
    return torch.sort(array).values


def diagonal_view(array, axis1=0, axis2=1):
    """Return a view of the array diagonal."""
    return torch.diagonal(array, 0, dim1=axis1, dim2=axis2)


def to_numpy(array):
    try:
        return array.cpu().numpy()
    except AttributeError:
        return array


def to_cpu(array):
    return array.cpu()


def to_gpu(array, device=None):
    return array


def isin(x, y):
    import numpy as np  # XXX
    np_result = np.isin(x.cpu().numpy(), y.cpu().numpy())
    return asarray(np_result, dtype=torch.bool, device=x.device)


def searchsorted(x, y):
    import numpy as np  # XXX
    np_result = np.searchsorted(x.cpu().numpy(), y.cpu().numpy())
    return asarray(np_result, dtype=torch.int64, device=x.device)


def flatnonzero(x):
    return torch.nonzero(torch.flatten(x), as_tuple=True)[0]


def asarray(x, dtype=None, device=None):
    if dtype is None:
        if isinstance(x, torch.Tensor):
            dtype = x.dtype
        if hasattr(x, "dtype") and hasattr(x.dtype, "name"):
            dtype = x.dtype.name
    if dtype is not None:
        dtype = _dtype_to_str(dtype)
        dtype = getattr(torch, dtype)
    if device is None and isinstance(x, torch.Tensor):
        device = x.device

    try:
        tensor = torch.as_tensor(x, dtype=dtype, device=device)
    except Exception:
        import numpy as np
        array = np.asarray(x, dtype=_dtype_to_str(dtype))
        tensor = torch.as_tensor(array, dtype=dtype, device=device)
    return tensor


def asarray_like(x, ref):
    return torch.as_tensor(x, dtype=ref.dtype, device=ref.device)


def norm(x, ord="fro", axis=None, keepdims=False):
    return torch.norm(x, p=ord, dim=axis, keepdim=keepdims)


def copy(x):
    return x.clone()


def transpose(a, axes=None):
    if axes is None:
        return a.t()
    else:
        return a.permute(*axes)


def max(*args, **kwargs):
    res = torch.max(*args, **kwargs)
    if isinstance(res, torch.Tensor):
        return res
    else:
        return res.values


def min(*args, **kwargs):
    res = torch.min(*args, **kwargs)
    if isinstance(res, torch.Tensor):
        return res
    else:
        return res.values


def zeros(shape, dtype="float32", device="cpu"):
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    return torch.zeros(shape, dtype=dtype, device=device)


def zeros_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in zeros_like."""
    if shape is None:
        shape = array.shape
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    if dtype is None:
        dtype = array.dtype
    if device is None:
        device = array.device
    return torch.zeros(shape, dtype=dtype, device=device, layout=array.layout)


def ones_like(array, shape=None, dtype=None, device=None):
    """Add a shape parameter in ones_like."""
    if shape is None:
        shape = array.shape
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    if dtype is None:
        dtype = array.dtype
    if device is None:
        device = array.device
    return torch.ones(shape, dtype=dtype, device=device, layout=array.layout)


def full_like(array, fill_value, shape=None, dtype=None, device=None):
    """Add a shape parameter in full_like."""
    if shape is None:
        shape = array.shape
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    if dtype is None:
        dtype = array.dtype
    if device is None:
        device = array.device
    return torch.full(shape, fill_value, dtype=dtype, device=device,
                      layout=array.layout)


def check_arrays(*all_inputs):
    """Change all inputs into Tensors (or list of Tensors) using the same
    precision and device as the first one. Some tensors can be None.
    """
    all_tensors = []
    all_tensors.append(asarray(all_inputs[0]))
    dtype = all_tensors[0].dtype
    device = all_tensors[0].device
    for tensor in all_inputs[1:]:
        if tensor is None:
            pass
        elif isinstance(tensor, list):
            tensor = [asarray(tt, dtype=dtype, device=device) for tt in tensor]
        else:
            tensor = asarray(tensor, dtype=dtype, device=device)
        all_tensors.append(tensor)
    return all_tensors


def svd(X, full_matrices=True):
    U, s, V = torch.svd(X, some=not full_matrices)
    return U, s, V.transpose(-2, -1)


try:
    eigh = torch.linalg.eigh
except AttributeError:
    # torch.__version__ < 1.8
    eigh = partial(torch.symeig, eigenvectors=True)



import numpy as np
import matplotlib.pyplot as plt
from himalaya.viz import plot_alphas_diagnostic


def test_smoke_viz():
    alphas = np.logspace(0, 5, 6)
    best_alphas = np.random.choice(np.logspace(0, 5, 6), 10)
    plot_alphas_diagnostic(best_alphas, alphas, ax=None)
    plot_alphas_diagnostic(best_alphas, alphas, ax=plt.gca())



import pytest

import numpy as np
import sklearn.linear_model
import sklearn.model_selection
import scipy.linalg

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal
from himalaya.scoring import r2_score

from himalaya.kernel_ridge import solve_multiple_kernel_ridge_random_search


def _create_dataset(backend):
    n_featuress = (100, 200)
    n_samples = 80
    n_targets = 4
    n_gammas = 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in n_featuress
    ]
    Ks = backend.stack([X @ X.T for X in Xs])

    ws = [
        backend.asarray(backend.randn(n_features, n_targets), backend.float64)
        for n_features in n_featuress
    ]
    Ys = backend.stack([X @ w for X, w in zip(Xs, ws)])
    Y = Ys.sum(0)

    gammas = backend.asarray(backend.rand(n_gammas, Ks.shape[0]),
                             backend.float64)
    gammas /= gammas.sum(1)[:, None]

    return Ks, Y, gammas, Xs


@pytest.mark.parametrize('local_alpha', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_random_search_local_alphah(
        backend, local_alpha):
    _test_solve_multiple_kernel_ridge_random_search(backend=backend,
                                                    local_alpha=local_alpha)


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_random_search_n_targets_batch(
        backend, n_targets_batch):
    _test_solve_multiple_kernel_ridge_random_search(
        backend=backend, n_targets_batch=n_targets_batch)


@pytest.mark.parametrize('n_alphas_batch', [None, 2])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_random_search_n_alphas_batch(
        backend, n_alphas_batch):
    _test_solve_multiple_kernel_ridge_random_search(
        backend=backend, n_alphas_batch=n_alphas_batch)


@pytest.mark.parametrize('return_weights', ['primal', 'dual'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_random_search_return_weights(
        backend, return_weights):
    _test_solve_multiple_kernel_ridge_random_search(
        backend=backend, return_weights=return_weights)


@pytest.mark.parametrize('diagonalize_method', ['eigh', 'svd'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_multiple_kernel_ridge_random_search_diagonalize_method(
        backend, diagonalize_method):
    _test_solve_multiple_kernel_ridge_random_search(
        backend=backend, diagonalize_method=diagonalize_method)


def _test_solve_multiple_kernel_ridge_random_search(
        backend, n_targets_batch=None, n_alphas_batch=None,
        return_weights="dual", diagonalize_method="eigh", local_alpha=True):
    backend = set_backend(backend)

    Ks, Y, gammas, Xs = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-3, 5, 9), Ks)
    n_targets = Y.shape[1]
    cv = sklearn.model_selection.check_cv(10)

    ############
    # run solver
    results = solve_multiple_kernel_ridge_random_search(
        Ks, Y, n_iter=gammas, alphas=alphas, score_func=r2_score, cv=cv,
        n_targets_batch=n_targets_batch, Xs=Xs, progress_bar=False,
        return_weights=return_weights, n_alphas_batch=n_alphas_batch,
        diagonalize_method=diagonalize_method, local_alpha=local_alpha)
    best_deltas, refit_weights, cv_scores = results

    #########################################
    # compare with sklearn.linear_model.Ridge
    if local_alpha:  # only compare when each target optimizes alpha
        test_scores = []
        for gamma in backend.sqrt(gammas):
            X = backend.concatenate([x * g for x, g in zip(Xs, gamma)], 1)
            for train, test in cv.split(X):
                for alpha in alphas:
                    model = sklearn.linear_model.Ridge(
                        alpha=backend.to_numpy(alpha), fit_intercept=False)
                    model = model.fit(backend.to_numpy(X[train]),
                                      backend.to_numpy(Y[train]))
                    predictions = backend.asarray_like(
                        model.predict(backend.to_numpy(X[test])), Y)
                    test_scores.append(r2_score(Y[test], predictions))

        test_scores = backend.stack(test_scores)
        test_scores = test_scores.reshape(len(gammas), cv.get_n_splits(),
                                          len(alphas), n_targets)
        test_scores_mean = backend.max(test_scores.mean(1), 1)
        assert_array_almost_equal(cv_scores, test_scores_mean, decimal=5)

    ######################
    # test refited_weights
    for tt in range(n_targets):
        gamma = backend.exp(best_deltas[:, tt])
        alpha = 1.0

        if return_weights == 'primal':
            # compare primal weights with sklearn.linear_model.Ridge
            X = backend.concatenate(
                [X * backend.sqrt(g) for X, g in zip(Xs, gamma)], 1)
            model = sklearn.linear_model.Ridge(fit_intercept=False,
                                               alpha=backend.to_numpy(alpha))
            w1 = model.fit(backend.to_numpy(X),
                           backend.to_numpy(Y[:, tt])).coef_
            w1 = np.split(w1, np.cumsum([X.shape[1] for X in Xs][:-1]), axis=0)
            w1 = [backend.asarray(w) for w in w1]
            w1_scaled = backend.concatenate(
                [w * backend.sqrt(g) for w, g, in zip(w1, gamma)])
            assert_array_almost_equal(w1_scaled, refit_weights[:, tt],
                                      decimal=5)

        elif return_weights == 'dual':
            # compare dual weights with scipy.linalg.solve
            Ks_64 = backend.asarray(Ks, dtype=backend.float64)
            gamma_64 = backend.asarray(gamma, dtype=backend.float64)
            K = backend.matmul(Ks_64.T, gamma_64).T
            reg = backend.asarray_like(np.eye(K.shape[0]), K) * alpha
            Y_64 = backend.asarray(Y, dtype=backend.float64)
            c1 = scipy.linalg.solve(backend.to_numpy(K + reg),
                                    backend.to_numpy(Y_64[:, tt]))
            c1 = backend.asarray_like(c1, K)
            assert_array_almost_equal(c1, refit_weights[:, tt], decimal=5)



import pytest
import numpy as np
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.ridge import ColumnTransformerNoStack
from himalaya.ridge import make_column_transformer_no_stack
from himalaya.ridge import GroupRidgeCV


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_transformer_all_columns(backend):
    backend = set_backend(backend)
    X = np.random.randn(10, 5)

    ct = ColumnTransformerNoStack([("name", StandardScaler(), slice(0, 5))])
    Xt = ct.fit_transform(X)
    assert len(Xt) == 1
    assert Xt[0].shape == (10, 5)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_transformer_passthrough(backend):
    backend = set_backend(backend)
    X = np.random.randn(10, 5)

    ct = ColumnTransformerNoStack([("name", "passthrough", slice(0, 5))])
    Xt = ct.fit_transform(X)
    assert len(Xt) == 1
    assert Xt[0].shape == (10, 5)
    assert_array_almost_equal(X, Xt[0])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_transformer_remainder(backend):
    backend = set_backend(backend)
    X = np.random.randn(10, 5)

    ct = ColumnTransformerNoStack([("name", "passthrough", [])],
                                  remainder="passthrough")
    Xt = ct.fit_transform(X)
    assert len(Xt) == 2
    assert Xt[0].shape == (10, 0)
    assert Xt[1].shape == (10, 5)
    assert_array_almost_equal(X, Xt[1])


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_transformer_multiple(backend):
    backend = set_backend(backend)
    X = np.random.randn(10, 5)

    ct = ColumnTransformerNoStack([
        ("name0", StandardScaler(), [0, 1]),
        ("name1", StandardScaler(with_mean=False), [2, 3]),
    ], remainder="passthrough")
    Xt = ct.fit_transform(X)
    assert len(Xt) == 3
    assert Xt[0].shape == (10, 2)
    assert Xt[1].shape == (10, 2)
    assert Xt[2].shape == (10, 1)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_make_column_transformer(backend):
    backend = set_backend(backend)

    trans = StandardScaler()
    ct = make_column_transformer_no_stack((trans, slice(0, 3)))

    assert isinstance(ct, ColumnTransformerNoStack)
    assert len(ct.transformers) == 1
    assert len(ct.transformers[0]) == 3
    assert ct.transformers[0][0] == "standardscaler"
    assert ct.transformers[0][1] == trans
    assert ct.transformers[0][2] == slice(0, 3)

    trans = StandardScaler()
    ct = make_column_transformer_no_stack((trans, slice(0, 3)),
                                          ("passthrough", [3, 4]))

    assert isinstance(ct, ColumnTransformerNoStack)
    assert len(ct.transformers) == 2
    assert len(ct.transformers[0]) == 3
    assert len(ct.transformers[1]) == 3
    assert ct.transformers[0][0] == "standardscaler"
    assert ct.transformers[0][1] == trans
    assert ct.transformers[0][2] == slice(0, 3)
    assert ct.transformers[1][0] == "passthrough"
    assert ct.transformers[1][1] == "passthrough"
    assert ct.transformers[1][2] == [3, 4]


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_column_transformer_in_pipeline(backend):
    backend = set_backend(backend)

    X = np.random.randn(10, 5)
    Y = np.random.randn(10, 3)

    ct = make_column_transformer_no_stack(
        (StandardScaler(), slice(0, 4)),
        (StandardScaler(), slice(4, 6)),
    )
    pipe = make_pipeline(
        ct,
        GroupRidgeCV(
            groups="input", solver_params=dict(n_iter=np.ones((1, 2)),
                                               progress_bar=False)))
    pipe.fit(X, Y)



"""Used for testing on GPU."""

from .torch import *  # noqa
import torch

try:
    torch.arange(1).cuda()
except AssertionError as error:
    import sys
    if "pytest" in sys.modules:  # if run through pytest
        import pytest
        pytest.skip("PyTorch not compiled with CUDA enabled.")
    raise AssertionError("PyTorch not compiled with CUDA enabled.") from error

from ._utils import _dtype_to_str

###############################################################################

name = "torch_cuda"


def randn(*args, **kwargs):
    return torch.randn(*args, **kwargs).cuda()


def rand(*args, **kwargs):
    return torch.rand(*args, **kwargs).cuda()


def asarray(x, dtype=None, device="cuda"):
    if dtype is None:
        if isinstance(x, torch.Tensor):
            dtype = x.dtype
        if hasattr(x, "dtype") and hasattr(x.dtype, "name"):
            dtype = x.dtype.name
    if dtype is not None:
        dtype = _dtype_to_str(dtype)
        dtype = getattr(torch, dtype)
    if device is None:
        if isinstance(x, torch.Tensor):
            device = x.device
        else:
            device = "cuda"
    try:
        tensor = torch.as_tensor(x, dtype=dtype, device=device)
    except Exception:
        import numpy as np
        array = np.asarray(x, dtype=_dtype_to_str(dtype))
        tensor = torch.as_tensor(array, dtype=dtype, device=device)
    return tensor


def zeros(shape, dtype="float32", device="cuda"):
    if isinstance(shape, int):
        shape = (shape, )
    if isinstance(dtype, str):
        dtype = getattr(torch, dtype)
    return torch.zeros(shape, dtype=dtype, device=device)


def to_cpu(array):
    return array.cpu()


def to_gpu(array, device="cuda"):
    return asarray(array, device=device)



import pytest

import numpy as np
import sklearn.linear_model
import scipy.linalg

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.ridge import RIDGE_SOLVERS


def _create_dataset(backend):
    n_samples, n_features, n_targets = 30, 10, 3

    X = backend.asarray(backend.randn(n_samples, n_features), backend.float64)
    Y = backend.asarray(backend.randn(n_samples, n_targets), backend.float64)
    weights = backend.asarray(backend.randn(n_features, n_targets),
                              backend.float64)

    return X, Y, weights


@pytest.mark.parametrize('solver_name', RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_kernel_ridge(solver_name, backend):
    backend = set_backend(backend)

    X, Y, weights = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-2, 5, 7), Y)

    solver = RIDGE_SOLVERS[solver_name]
    XTX = X.T @ X
    XTY = X.T @ Y

    for alpha in alphas:
        alpha = backend.full_like(Y, fill_value=alpha, shape=Y.shape[1])
        b2 = solver(X, Y, alpha=alpha, fit_intercept=False)
        assert b2.shape == (X.shape[1], Y.shape[1])

        n_features, n_targets = weights.shape
        for ii in range(n_targets):
            # compare primal coefficients with scipy.linalg.solve
            reg = backend.asarray_like(np.eye(n_features), Y) * alpha[ii]
            b1 = scipy.linalg.solve(backend.to_numpy(XTX + reg),
                                    backend.to_numpy(XTY[:, ii]))
            assert_array_almost_equal(b1, b2[:, ii], decimal=6)

            # compare predictions with sklearn.linear_model.Ridge
            prediction = backend.matmul(X, b2[:, ii])
            model = sklearn.linear_model.Ridge(
                alpha=backend.to_numpy(alpha[ii]),
                max_iter=1000, tol=1e-6, fit_intercept=False)
            model.fit(backend.to_numpy(X), backend.to_numpy(Y[:, ii]))
            prediction_sklearn = model.predict(backend.to_numpy(X))
            assert_array_almost_equal(prediction, prediction_sklearn,
                                      decimal=6)

            assert_array_almost_equal(model.coef_, b2[:, ii], decimal=5)


@pytest.mark.parametrize('solver_name', RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_kernel_ridge_intercept(solver_name, backend):
    backend = set_backend(backend)

    X, Y, weights = _create_dataset(backend)
    Y += 100
    X += 10
    alphas = backend.asarray_like(backend.logspace(-2, 5, 7), Y)

    solver = RIDGE_SOLVERS[solver_name]

    for alpha in alphas:
        alpha = backend.full_like(Y, fill_value=alpha, shape=Y.shape[1])
        b2, i2 = solver(X, Y, alpha=alpha, fit_intercept=True)
        assert b2.shape == (X.shape[1], Y.shape[1])
        assert i2.shape == (Y.shape[1], )

        n_features, n_targets = weights.shape
        for ii in range(n_targets):

            # compare predictions with sklearn.linear_model.Ridge
            prediction = backend.matmul(X, b2[:, ii]) + i2[ii]
            model = sklearn.linear_model.Ridge(
                alpha=backend.to_numpy(alpha[ii]),
                max_iter=1000, tol=1e-6, fit_intercept=True)
            model.fit(backend.to_numpy(X), backend.to_numpy(Y[:, ii]))
            prediction_sklearn = model.predict(backend.to_numpy(X))
            assert_array_almost_equal(prediction, prediction_sklearn,
                                      decimal=5)

            assert_array_almost_equal(model.coef_, b2[:, ii], decimal=5)



import pytest

import numpy as np
import sklearn.linear_model
import sklearn.model_selection

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal
from himalaya.scoring import r2_score

from himalaya.ridge import solve_group_ridge_random_search


def _create_dataset(backend):
    n_featuress = (10, 20)
    n_samples = 80
    n_targets = 4
    n_gammas = 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in n_featuress
    ]

    ws = [
        backend.asarray(backend.randn(n_features, n_targets), backend.float64)
        for n_features in n_featuress
    ]
    Ys = backend.stack([X @ w for X, w in zip(Xs, ws)])
    Y = Ys.sum(0)
    Y += backend.asarray(backend.randn(*Y.shape), backend.float64)

    gammas = backend.asarray(backend.rand(n_gammas, len(Xs)), backend.float64)
    gammas /= gammas.sum(1)[:, None]

    return Xs, Y, gammas


@pytest.mark.parametrize('n_targets_batch', [None, 3])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_group_ridge_random_search_n_targets_batch(
        backend, n_targets_batch):
    _test_solve_group_ridge_random_search(backend=backend,
                                           n_targets_batch=n_targets_batch)


@pytest.mark.parametrize('n_alphas_batch', [None, 2])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_group_ridge_random_search_n_alphas_batch(
        backend, n_alphas_batch):
    _test_solve_group_ridge_random_search(backend=backend,
                                           n_alphas_batch=n_alphas_batch)


def _test_solve_group_ridge_random_search(backend, n_targets_batch=None,
                                           n_alphas_batch=None):
    backend = set_backend(backend)

    Xs, Y, gammas = _create_dataset(backend)
    alphas = backend.asarray_like(backend.logspace(-3, 5, 9), Xs[0])
    n_targets = Y.shape[1]
    cv = sklearn.model_selection.check_cv(10)

    ############
    # run solver
    results = solve_group_ridge_random_search(
        Xs, Y, n_iter=gammas, alphas=alphas, score_func=r2_score, cv=cv,
        n_targets_batch=n_targets_batch, progress_bar=False,
        return_weights=True, n_alphas_batch=n_alphas_batch,
        diagonalize_method="svd")
    best_deltas, refit_weights, cv_scores = results

    #########################################
    # compare with sklearn.linear_model.Ridge
    test_scores = []
    for gamma in gammas:
        X = backend.concatenate(
            [x * backend.sqrt(g) for x, g in zip(Xs, gamma)], 1)
        for train, test in cv.split(X):
            for alpha in alphas:
                model = sklearn.linear_model.Ridge(
                    alpha=backend.to_numpy(alpha), fit_intercept=False)
                model = model.fit(backend.to_numpy(X[train]),
                                  backend.to_numpy(Y[train]))
                predictions = backend.asarray_like(
                    model.predict(backend.to_numpy(X[test])), ref=Y)
                test_scores.append(r2_score(Y[test], predictions))

    test_scores = backend.stack(test_scores)
    test_scores = test_scores.reshape(len(gammas), cv.get_n_splits(),
                                      len(alphas), n_targets)
    test_scores_mean = backend.max(test_scores.mean(1), 1)
    assert_array_almost_equal(cv_scores, test_scores_mean, decimal=5)

    ######################
    # test refited_weights
    for tt in range(n_targets):
        gamma = backend.exp(best_deltas[:, tt])
        alpha = 1.0

        # compare primal weights with sklearn.linear_model.Ridge
        X = backend.concatenate(
            [X * backend.sqrt(g) for X, g in zip(Xs, gamma)], 1)
        model = sklearn.linear_model.Ridge(fit_intercept=False,
                                           alpha=backend.to_numpy(alpha))
        w1 = model.fit(backend.to_numpy(X), backend.to_numpy(Y[:, tt])).coef_
        w1 = np.split(w1, np.cumsum([X.shape[1] for X in Xs][:-1]), axis=0)
        w1 = [backend.asarray(w) for w in w1]
        w1_scaled = backend.concatenate(
            [w * backend.sqrt(g) for w, g, in zip(w1, gamma)])
        assert_array_almost_equal(w1_scaled, refit_weights[:, tt], decimal=5)



import numpy as np
import sklearn
import pytest

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.validation import _assert_all_finite


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_suppress_validation(backend):
    backend = set_backend(backend)
    X = backend.asarray([0, np.inf])
    with pytest.raises(ValueError):
        _assert_all_finite(X, True)
    sklearn.set_config(assume_finite=True)
    _assert_all_finite(X, True)
    sklearn.set_config(assume_finite=False)
    with pytest.raises(ValueError):
        _assert_all_finite(X, True)



from functools import partial
import pytest

import numpy as np
import sklearn.linear_model
import scipy.linalg

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.kernel_ridge import WEIGHTED_KERNEL_RIDGE_SOLVERS
from himalaya.kernel_ridge import KERNEL_RIDGE_SOLVERS
from himalaya.kernel_ridge._solvers import _weighted_kernel_ridge_gradient

KERNEL_RIDGE_SOLVERS['eigenvalues_svd'] = partial(
    KERNEL_RIDGE_SOLVERS['eigenvalues'], method="svd")


def _create_dataset(backend, intercept):
    n_samples, n_targets = 30, 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in [100, 200]
    ]
    if intercept:
        Xs[0] += 1.1
        Xs[1] += 1.2

    Ks = backend.stack([backend.matmul(X, X.T) for X in Xs])
    Y = backend.asarray(backend.randn(n_samples, n_targets), backend.float64)
    dual_weights = backend.asarray(backend.randn(n_samples, n_targets),
                                   backend.float64)
    exp_deltas = backend.asarray(backend.rand(Ks.shape[0], n_targets),
                                 backend.float64)
    deltas = backend.log(exp_deltas)

    if intercept:
        Y += -8

    return Xs, Ks, Y, deltas, dual_weights


@pytest.mark.parametrize("double_K", [False, True])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_weighted_kernel_ridge_gradient(backend, double_K):
    backend = set_backend(backend)

    _, Ks, Y, deltas, dual_weights = _create_dataset(backend, intercept=False)
    exp_deltas = backend.exp(deltas)
    alpha = 1.

    n_targets = Y.shape[1]
    grad = backend.zeros_like(dual_weights, dtype=backend.float64)
    func = backend.zeros_like(dual_weights, dtype=backend.float64,
                              shape=(n_targets))
    for tt in range(n_targets):
        K = backend.sum(
            backend.stack([K * g for K, g in zip(Ks, exp_deltas[:, tt])]), 0)
        grad[:, tt] = (backend.matmul(K, dual_weights[:, tt]) - Y[:, tt] +
                       alpha * dual_weights[:, tt])

        pred = backend.matmul(K, dual_weights[:, tt])
        func[tt] = backend.sum((pred - Y[:, tt]) ** 2, 0)
        func[tt] += alpha * (dual_weights[:, tt] @ K @ dual_weights[:, tt])

        if double_K:
            grad[:, tt] = backend.matmul(K.T, grad[:, tt])

    ########################
    grad2, func2 = _weighted_kernel_ridge_gradient(Ks, Y, dual_weights,
                                                   exp_deltas=exp_deltas,
                                                   alpha=alpha,
                                                   double_K=double_K,
                                                   return_objective=True)
    assert_array_almost_equal(grad, grad2)
    assert_array_almost_equal(func, func2)


@pytest.mark.parametrize('solver_name', WEIGHTED_KERNEL_RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_weighted_kernel_ridge(solver_name, backend):
    backend = set_backend(backend)

    solver = WEIGHTED_KERNEL_RIDGE_SOLVERS[solver_name]
    decimal = 1 if solver_name == "neumann_series" else 3

    Xs, Ks, Y, deltas, dual_weights = _create_dataset(backend, intercept=False)
    exp_deltas = backend.exp(deltas)

    for alpha in backend.asarray_like(backend.logspace(-2, 3, 7), Ks):
        c2 = solver(Ks, Y, deltas, alpha=alpha, max_iter=3000, tol=1e-6)

        n_targets = Y.shape[1]
        for ii in range(n_targets):
            # compare dual coefficients with scipy.linalg.solve
            K = backend.matmul(Ks.T, exp_deltas[:, ii]).T
            reg = backend.asarray_like(np.eye(K.shape[0]), K) * alpha
            c1 = scipy.linalg.solve(backend.to_numpy(K + reg),
                                    backend.to_numpy(Y[:, ii]))
            assert_array_almost_equal(c1, c2[:, ii], decimal=decimal)

            if solver_name != "neumann_series":
                # compare predictions with sklearn.linear_model.Ridge
                X_scaled = backend.concatenate([
                    t * backend.sqrt(g) for t, g in zip(Xs, exp_deltas[:, ii])
                ], 1)
                prediction = backend.matmul(K, c2[:, ii])
                model = sklearn.linear_model.Ridge(
                    alpha=backend.to_numpy(alpha), solver="lsqr",
                    max_iter=1000, tol=1e-6, fit_intercept=False)
                model.fit(backend.to_numpy(X_scaled),
                          backend.to_numpy(Y[:, ii]))
                prediction_sklearn = model.predict(backend.to_numpy(X_scaled))
                assert_array_almost_equal(prediction, prediction_sklearn,
                                          decimal=decimal)


@pytest.mark.parametrize('solver_name', WEIGHTED_KERNEL_RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_weighted_kernel_ridge_intercept(solver_name, backend):
    backend = set_backend(backend)

    solver = WEIGHTED_KERNEL_RIDGE_SOLVERS[solver_name]
    if solver_name == "neumann_series":
        pytest.skip()

    Xs, Ks, Y, deltas, dual_weights = _create_dataset(backend, intercept=True)
    exp_deltas = backend.exp(deltas)

    # torch with cuda has more limited precision in mean
    decimal = 1 if backend.name == "torch_cuda" else 5

    for alpha in backend.asarray_like(backend.logspace(-2, 3, 7), Ks):
        c2, i2 = solver(Ks, Y, deltas, alpha=alpha, max_iter=100, tol=1e-6,
                        fit_intercept=True)

        n_targets = Y.shape[1]
        for ii in range(n_targets):
            K = backend.matmul(Ks.T, exp_deltas[:, ii]).T
            # compare predictions with sklearn.linear_model.Ridge
            X_scaled = backend.concatenate(
                [t * backend.sqrt(g) for t, g in zip(Xs, exp_deltas[:, ii])],
                1)
            prediction = backend.matmul(K, c2[:, ii]) + i2[ii]
            model = sklearn.linear_model.Ridge(alpha=backend.to_numpy(alpha),
                                               solver="lsqr", max_iter=1000,
                                               tol=1e-6, fit_intercept=True)
            model.fit(backend.to_numpy(X_scaled), backend.to_numpy(Y[:, ii]))
            prediction_sklearn = model.predict(backend.to_numpy(X_scaled))
            assert_array_almost_equal(prediction, prediction_sklearn,
                                      decimal=decimal)


@pytest.mark.parametrize('solver_name', KERNEL_RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_kernel_ridge(solver_name, backend):
    backend = set_backend(backend)

    Xs, Ks, Y, deltas, dual_weights = _create_dataset(backend, intercept=False)
    alphas = backend.asarray_like(backend.logspace(-2, 5, 7), Ks)

    solver = KERNEL_RIDGE_SOLVERS[solver_name]

    deltas = deltas[:, 0]
    exp_deltas = backend.exp(deltas)
    K = backend.matmul(Ks.T, exp_deltas).T

    for alpha in alphas:
        alpha = backend.full_like(K, fill_value=alpha, shape=Y.shape[1])
        if "eigenvalues" in solver_name:
            c2 = solver(K, Y, alpha=alpha)
        else:
            c2 = solver(K, Y, alpha=alpha, max_iter=3000, tol=1e-6)

        n_targets = Y.shape[1]
        for ii in range(n_targets):
            # compare dual coefficients with scipy.linalg.solve
            reg = backend.asarray_like(np.eye(K.shape[0]), K) * alpha[ii]
            c1 = scipy.linalg.solve(backend.to_numpy(K + reg),
                                    backend.to_numpy(Y[:, ii]))
            assert_array_almost_equal(c1, c2[:, ii], decimal=3)

            # compare predictions with sklearn.linear_model.Ridge
            X_scaled = backend.concatenate(
                [t * backend.sqrt(g) for t, g in zip(Xs, exp_deltas)], 1)
            prediction = backend.matmul(K, c2[:, ii])
            model = sklearn.linear_model.Ridge(
                alpha=backend.to_numpy(alpha[ii]), solver="lsqr",
                max_iter=1000, tol=1e-6, fit_intercept=False)
            model.fit(backend.to_numpy(X_scaled), backend.to_numpy(Y[:, ii]))
            prediction_sklearn = model.predict(backend.to_numpy(X_scaled))
            assert_array_almost_equal(prediction, prediction_sklearn,
                                      decimal=5)


@pytest.mark.parametrize('solver_name', KERNEL_RIDGE_SOLVERS)
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_solve_kernel_ridge_intercept(solver_name, backend):
    backend = set_backend(backend)

    Xs, Ks, Y, deltas, dual_weights = _create_dataset(backend, intercept=True)

    alphas = backend.asarray_like(backend.logspace(-2, 5, 7), Ks)

    solver = KERNEL_RIDGE_SOLVERS[solver_name]

    deltas = deltas[:, 0]
    exp_deltas = backend.exp(deltas)
    K = backend.matmul(Ks.T, exp_deltas).T

    # torch with cuda has more limited precision in mean
    decimal = 1 if backend.name == "torch_cuda" else 5

    for alpha in alphas:
        alpha = backend.full_like(K, fill_value=alpha, shape=Y.shape[1])
        if "eigenvalues" in solver_name:
            c2, i2 = solver(K, Y, alpha=alpha, fit_intercept=True)
        else:
            c2, i2 = solver(K, Y, alpha=alpha, fit_intercept=True,
                            max_iter=100, tol=1e-5)

        n_targets = Y.shape[1]
        for ii in range(n_targets):
            # compare predictions with sklearn.linear_model.Ridge
            X_scaled = backend.concatenate(
                [t * backend.sqrt(g) for t, g in zip(Xs, exp_deltas)], 1)
            prediction = backend.matmul(K, c2[:, ii]) + i2[ii]
            model = sklearn.linear_model.Ridge(
                alpha=backend.to_numpy(alpha[ii]), solver="lsqr",
                max_iter=1000, tol=1e-6, fit_intercept=True)
            model.fit(backend.to_numpy(X_scaled), backend.to_numpy(Y[:, ii]))
            prediction_sklearn = model.predict(backend.to_numpy(X_scaled))
            assert_array_almost_equal(prediction, prediction_sklearn,
                                      decimal=decimal)



import warnings

import pytest
import sklearn.kernel_ridge
import sklearn.utils.estimator_checks

from himalaya.backend import set_backend
from himalaya.backend import get_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.ridge import Ridge
from himalaya.ridge import RidgeCV
from himalaya.kernel_ridge import KernelRidge
from himalaya.kernel_ridge import KernelRidgeCV
from himalaya.kernel_ridge import MultipleKernelRidgeCV
from himalaya.kernel_ridge import WeightedKernelRidge


def _create_dataset(backend):
    n_samples, n_targets = 30, 3

    Xs = [
        backend.asarray(backend.randn(n_samples, n_features), backend.float64)
        for n_features in [100, 200]
    ]
    Ks = backend.stack([backend.matmul(X, X.T) for X in Xs])
    Y = backend.asarray(backend.randn(n_samples, n_targets), backend.float64)

    return Xs, Ks, Y


@pytest.mark.parametrize('kernel', [
    'linear', 'polynomial', 'poly', 'rbf', 'sigmoid', 'cosine', 'precomputed'
])
@pytest.mark.parametrize('multitarget', [True, False])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_vs_scikit_learn(backend, multitarget, kernel):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    if not multitarget:
        Y = Y[:, 0]

    if kernel == "precomputed":
        X = Ks[0]
    else:
        X = Xs[0]

    for alpha in backend.asarray_like(backend.logspace(0, 3, 7), Ks):
        model = KernelRidge(alpha=alpha, kernel=kernel)
        model.fit(X, Y)

        reference = sklearn.kernel_ridge.KernelRidge(
            alpha=backend.to_numpy(alpha), kernel=kernel)
        reference.fit(backend.to_numpy(X), backend.to_numpy(Y))

        assert model.dual_coef_.shape == Y.shape
        assert_array_almost_equal(model.dual_coef_, reference.dual_coef_)
        assert_array_almost_equal(model.predict(X),
                                  reference.predict(backend.to_numpy(X)))
        assert_array_almost_equal(
            model.score(X, Y).mean(),
            reference.score(backend.to_numpy(X), backend.to_numpy(Y)))


@pytest.mark.parametrize('fit_intercept', [False, True])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_vs_ridge(backend, fit_intercept):
    # useful to test the intercept as well
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)
    X = Xs[0]
    if fit_intercept:
        Y += 10
        X += 1

    # torch with cuda has more limited precision in mean
    decimal = 3 if backend.name == "torch_cuda" else 6

    for alpha in backend.asarray_like(backend.logspace(0, 3, 7), X):
        model = KernelRidge(alpha=alpha, fit_intercept=fit_intercept)
        model.fit(X, Y)
        reference = Ridge(alpha=alpha, fit_intercept=fit_intercept)
        reference.fit(backend.to_numpy(X), backend.to_numpy(Y))

        assert_array_almost_equal(model.predict(X), reference.predict(X),
                                  decimal=decimal)


@pytest.mark.parametrize('fit_intercept', [False, True])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_cv_vs_ridge_cv(backend, fit_intercept):
    # useful to test the intercept as well
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)
    X = Xs[0]
    if fit_intercept:
        Y += 10
        Xs[0] += 1
    alphas = backend.asarray_like(backend.logspace(-2, 3, 21), Y)  # XXX

    # torch with cuda has more limited precision in mean
    decimal = 4 if backend.name == "torch_cuda" else 6

    model = KernelRidgeCV(alphas=alphas, fit_intercept=fit_intercept)
    model.fit(X, Y)
    reference = RidgeCV(alphas=alphas, fit_intercept=fit_intercept)
    reference.fit(X, Y)

    assert_array_almost_equal(model.best_alphas_, reference.best_alphas_,
                              decimal=5)
    assert_array_almost_equal(model.predict(X), reference.predict(X),
                              decimal=decimal)


@pytest.mark.parametrize(
    'kernel', ['linear', 'polynomial', 'poly', 'rbf', 'sigmoid', 'cosine'])
@pytest.mark.parametrize('format', ['coo', 'csr', 'csc'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_vs_scikit_learn_sparse(backend, kernel, format):
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)

    try:
        import scipy.sparse
    except ImportError:
        pytest.skip("Scipy is not installed.")

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', scipy.sparse.SparseEfficiencyWarning)
        X = scipy.sparse.rand(*Xs[0].shape, density=0.1, format=format)

    for alpha in backend.asarray_like(backend.logspace(0, 3, 7), Y):
        model = KernelRidge(alpha=alpha, kernel=kernel)
        model.fit(X, Y)

        reference = sklearn.kernel_ridge.KernelRidge(
            alpha=backend.to_numpy(alpha), kernel=kernel)
        reference.fit(X, backend.to_numpy(Y))

        assert model.dual_coef_.shape == Y.shape
        assert_array_almost_equal(model.dual_coef_, reference.dual_coef_)
        assert_array_almost_equal(model.predict(X), reference.predict(X))
        assert_array_almost_equal(
            model.score(X, Y).mean(), reference.score(X, backend.to_numpy(Y)))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_precomputed(backend):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    for alpha in backend.asarray_like(backend.logspace(-2, 3, 7), Ks):
        model_1 = KernelRidge(alpha=alpha, kernel="linear")
        model_1.fit(Xs[0], Y)
        model_2 = KernelRidge(alpha=alpha, kernel="precomputed")
        model_2.fit(Ks[0], Y)

        assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
        assert_array_almost_equal(model_1.predict(Xs[0]),
                                  model_2.predict(Ks[0]))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_get_primal_coef(backend):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model = KernelRidge(kernel="linear")
    model.fit(Xs[0], Y)
    primal_coef = model.get_primal_coef()
    predictions_primal = Xs[0] @ backend.asarray(primal_coef)
    predictions_dual = model.predict(Xs[0])
    assert_array_almost_equal(predictions_primal, predictions_dual)

    model = KernelRidge(kernel="precomputed")
    model.fit(Ks[0], Y)
    primal_coef = model.get_primal_coef(X_fit=Xs[0])
    predictions_primal = Xs[0] @ backend.asarray(primal_coef)
    predictions_dual = model.predict(Ks[0])
    assert_array_almost_equal(predictions_primal, predictions_dual)

    model = KernelRidge(kernel="precomputed")
    model.fit(Ks[0], Y)
    with pytest.raises(ValueError):
        model.get_primal_coef(X_fit=None)

    model = KernelRidge(kernel="poly")
    model.fit(Xs[0], Y)
    with pytest.raises(ValueError):
        model.get_primal_coef()


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_weighted_kernel_ridge_get_primal_coef(backend):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model = WeightedKernelRidge(kernels="precomputed")
    model.fit(Ks, Y)

    with pytest.raises(ValueError):
        primal_coef = model.get_primal_coef(Xs_fit=None)

    primal_coef = model.get_primal_coef(Xs_fit=Xs)
    predictions_primal = backend.stack(
        [X @ backend.asarray(w) for X, w in zip(Xs, primal_coef)]).sum(0)
    predictions_dual = model.predict(Ks)
    assert_array_almost_equal(predictions_primal, predictions_dual)


@pytest.mark.parametrize(
    'solver', ['eigenvalues', 'conjugate_gradient', 'gradient_descent'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_solvers(solver, backend):
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)

    kernel = "linear"
    X = Xs[0]

    if solver == "eigenvalues":
        solver_params = dict()
    elif solver == "conjugate_gradient":
        solver_params = dict(max_iter=300, tol=1e-6)
    elif solver == "gradient_descent":
        solver_params = dict(max_iter=300, tol=1e-6)

    for alpha in backend.asarray_like(backend.logspace(0, 3, 7), Y):
        model = KernelRidge(alpha=alpha, kernel=kernel, solver=solver,
                            solver_params=solver_params)
        model.fit(X, Y)

        reference = sklearn.kernel_ridge.KernelRidge(
            alpha=backend.to_numpy(alpha), kernel=kernel)
        reference.fit(backend.to_numpy(X), backend.to_numpy(Y))

        assert model.dual_coef_.shape == Y.shape
        assert_array_almost_equal(model.dual_coef_, reference.dual_coef_)
        assert_array_almost_equal(model.predict(X),
                                  reference.predict(backend.to_numpy(X)),
                                  decimal=5)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_wrong_solver(backend):
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)
    X = Xs[0]

    model = KernelRidge(solver="wrong")
    with pytest.raises(ValueError, match="Unknown solver"):
        model.fit(X, Y)


@pytest.mark.parametrize('solver', ['eigenvalues'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_cv_precomputed(backend, solver):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model_1 = KernelRidgeCV(kernel="linear")
    model_1.fit(Xs[0], Y)
    model_2 = KernelRidgeCV(kernel="precomputed")
    model_2.fit(Ks[0], Y)

    assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
    assert_array_almost_equal(model_1.predict(Xs[0]), model_2.predict(Ks[0]))


@pytest.mark.parametrize('solver', ['random_search', 'hyper_gradient'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_multiple_kernel_ridge_cv_precomputed(backend, solver):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    if solver == "random_search":
        kwargs = dict(solver="random_search", random_state=0,
                      solver_params=dict(n_iter=2, progress_bar=False))
    elif solver == "hyper_gradient":
        kwargs = dict(solver="hyper_gradient", random_state=0,
                      solver_params=dict(max_iter=2, progress_bar=False))

    model_1 = MultipleKernelRidgeCV(kernels=["linear"], **kwargs)
    model_1.fit(Xs[0], Y)
    model_2 = MultipleKernelRidgeCV(kernels="precomputed", **kwargs)
    model_2.fit(Ks[0][None], Y)

    assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
    assert_array_almost_equal(model_1.predict(Xs[0]),
                              model_2.predict(Ks[0][None]))


@pytest.mark.parametrize('solver', ['conjugate_gradient', 'gradient_descent'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_weighted_kernel_ridge_precomputed(backend, solver):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model_1 = WeightedKernelRidge(kernels=["linear"])
    model_1.fit(Xs[0], Y)
    model_2 = WeightedKernelRidge(kernels="precomputed")
    model_2.fit(Ks[0][None], Y)

    assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
    assert_array_almost_equal(model_1.predict(Xs[0]),
                              model_2.predict(Ks[0][None]))


@pytest.mark.parametrize('Estimator',
                         [WeightedKernelRidge, MultipleKernelRidgeCV])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_weighted_kernel_ridge_split_predict(backend, Estimator):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    # multiple targets
    model = Estimator(kernels="precomputed")
    model.fit(Ks, Y)
    Y_pred = model.predict(Ks)
    Y_pred_split = model.predict(Ks, split=True)
    assert Y_pred.shape == Y.shape
    assert_array_almost_equal(Y_pred, Y_pred_split.sum(0))

    # single targets
    model = Estimator(kernels="precomputed")
    model.fit(Ks, Y[:, 0])
    Y_pred = model.predict(Ks)
    Y_pred_split = model.predict(Ks, split=True)
    assert Y_pred.shape == Y[:, 0].shape
    assert_array_almost_equal(Y_pred, Y_pred_split.sum(0))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_duplicate_solver_parameters(backend):
    backend = set_backend(backend)
    Xs, _, Y = _create_dataset(backend)

    model = KernelRidge(solver_params=dict(alpha=1))
    with pytest.raises(ValueError):
        model.fit(Xs[0], Y)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_cv_predict(backend):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    n_iter = backend.ones_like(Ks, shape=(1, Ks.shape[0]))
    alphas = backend.logspace(1, 2, 3)

    model_0 = KernelRidgeCV(kernel="precomputed",
                            alphas=alphas).fit(Ks.sum(0), Y)
    model_1 = MultipleKernelRidgeCV(
        kernels="precomputed", solver_params=dict(n_iter=n_iter,
                                                  alphas=alphas)).fit(Ks, Y)

    assert_array_almost_equal(model_0.predict(Ks.sum(0)), model_1.predict(Ks))


@pytest.mark.parametrize('solver', ['eigenvalues'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_kernel_ridge_cv_Y_in_cpu(backend, solver):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model_1 = KernelRidgeCV(solver=solver, Y_in_cpu=True)
    model_1.fit(Xs[0], Y)
    model_2 = KernelRidgeCV(solver=solver, Y_in_cpu=False)
    model_2.fit(Xs[0], Y)

    assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
    assert_array_almost_equal(model_1.predict(Xs[0]), model_2.predict(Xs[0]))


@pytest.mark.parametrize('solver', ['random_search', 'hyper_gradient'])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_multiple_kernel_ridge_cv_Y_in_cpu(backend, solver):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    model_1 = MultipleKernelRidgeCV(kernels="precomputed", solver=solver,
                                    Y_in_cpu=True, random_state=0)
    model_1.fit(Ks, Y)
    model_2 = MultipleKernelRidgeCV(kernels="precomputed", solver=solver,
                                    Y_in_cpu=False, random_state=0)
    model_2.fit(Ks, Y)

    assert_array_almost_equal(model_1.dual_coef_, model_2.dual_coef_)
    assert_array_almost_equal(model_1.predict(Ks), model_2.predict(Ks))


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_weighted_kernel_ridge_cv_array_deltas(backend):
    backend = set_backend(backend)
    Xs, Ks, Y = _create_dataset(backend)

    # correct deltas array
    deltas = backend.zeros_like(Ks, shape=(len(Ks), ))
    model_1 = WeightedKernelRidge(kernels="precomputed", deltas=deltas,
                                  random_state=0)
    model_1.fit(Ks, Y)

    # wrong number of kernels
    deltas = backend.zeros_like(Ks, shape=(len(Ks) + 1, ))
    model_1 = WeightedKernelRidge(kernels="precomputed", deltas=deltas,
                                  random_state=0)
    with pytest.raises(ValueError, match="Inconsistent number of kernels"):
        model_1.fit(Ks, Y)

    # wrong number of targets
    deltas = backend.zeros_like(Ks, shape=(len(Ks), Y.shape[1] + 1))
    model_1 = WeightedKernelRidge(kernels="precomputed", deltas=deltas,
                                  random_state=0)
    with pytest.raises(ValueError, match="Inconsistent number of targets"):
        model_1.fit(Ks, Y)


###############################################################################
###############################################################################
###############################################################################
# scikit-learn.utils.estimator_checks


class KernelRidge_(KernelRidge):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        from himalaya.validation import check_array
        from himalaya.scoring import r2_score
        backend = get_backend()

        y_pred = super().predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.dual_coef_.ndim)

        if y_true.ndim == 1:
            return backend.to_numpy(
                r2_score(y_true[:, None], y_pred[:, None])[0])
        else:
            return backend.to_numpy(r2_score(y_true, y_pred))


class KernelRidgeCV_(KernelRidgeCV):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, alphas=(0.1, 1), kernel="linear", kernel_params=None,
                 solver="eigenvalues", solver_params=None, cv=2):
        super().__init__(alphas=alphas, kernel=kernel,
                         kernel_params=kernel_params, solver=solver,
                         solver_params=solver_params, cv=cv)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        from himalaya.validation import check_array
        from himalaya.scoring import r2_score
        backend = get_backend()

        y_pred = super().predict(X)
        y_true = check_array(y, dtype=self.dtype_, ndim=self.dual_coef_.ndim)

        if y_true.ndim == 1:
            return backend.to_numpy(
                r2_score(y_true[:, None], y_pred[:, None])[0])
        else:
            return backend.to_numpy(r2_score(y_true, y_pred))


class MultipleKernelRidgeCV_(MultipleKernelRidgeCV):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, kernels=("linear", "polynomial"), kernels_params=None,
                 solver="hyper_gradient", solver_params=None, cv=2,
                 random_state=None):
        super().__init__(kernels=kernels, kernels_params=kernels_params,
                         solver=solver, solver_params=solver_params, cv=cv,
                         random_state=random_state)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        backend = get_backend()
        return backend.to_numpy(super().score(X, y))


class WeightedKernelRidge_(WeightedKernelRidge):
    """Cast predictions to numpy arrays, to be used in scikit-learn tests.

    Used for testing only.
    """

    def __init__(self, alpha=1., deltas="zeros",
                 kernels=("linear", "polynomial"), kernels_params=None,
                 solver="conjugate_gradient", solver_params=None,
                 random_state=None):
        super().__init__(alpha=alpha, deltas=deltas, kernels=kernels,
                         kernels_params=kernels_params, solver=solver,
                         solver_params=solver_params,
                         random_state=random_state)

    def predict(self, X):
        backend = get_backend()
        return backend.to_numpy(super().predict(X))

    def score(self, X, y):
        backend = get_backend()
        return backend.to_numpy(super().score(X, y))


@sklearn.utils.estimator_checks.parametrize_with_checks([
    KernelRidge_(),
    KernelRidgeCV_(),
    MultipleKernelRidgeCV_(),
    WeightedKernelRidge_(),
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_check_estimator(estimator, check, backend):
    backend = set_backend(backend)
    check(estimator)



from ..backend import get_backend
from ..progress_bar import bar


def predict_weighted_kernel_ridge(Ks, dual_weights, deltas, split=False):
    """
    Compute predictions, typically on a test set.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples_test, n_samples_train)
        Test kernels.
    dual_weights : array of shape (n_samples_train, n_targets)
        Dual weights of the kernel ridge model.
    deltas : array of shape (n_kernels, n_targets)
        Log kernel weights for each target.
    split : bool
        If True, the predictions is split across kernels.

    Returns
    -------
    Y_hat : array of shape (n_samples_test, n_targets) or \
            (n_kernels, n_samples_test, n_targets) (if split is True)
        Predicted values.
    """
    backend = get_backend()

    Ks, dual_weights, deltas = backend.check_arrays(Ks, dual_weights, deltas)
    chi = backend.matmul(Ks, dual_weights)
    split_predictions = backend.exp(deltas[:, None, :]) * chi
    if split:
        Y_hat = split_predictions
    else:
        Y_hat = split_predictions.sum(0)

    return Y_hat


def predict_and_score_weighted_kernel_ridge(Ks, dual_weights, deltas, Y,
                                            score_func, split=False,
                                            n_targets_batch=None,
                                            progress_bar=False):
    """
    Compute predictions, typically on a test set, and compute the score.

    Parameters
    ----------
    Ks : array of shape (n_kernels, n_samples_test, n_samples_train)
        Input kernels.
    dual_weights : array of shape (n_samples_train, n_targets)
        Dual weights of the kernel ridge model.
    deltas : array of shape (n_kernels, n_targets)
        Log kernel weights for each target.
    Y : array of shape (n_samples_test, n_targets)
        Target data.
    score_func : callable
        Function used to compute the score of predictions.
    split : bool
        If True, the predictions is split across kernels.
    n_targets_batch : int or None
        Size of the batch for computing predictions. Used for memory reasons.
        If None, uses all n_targets at once.
    progress_bar : bool
        If True, display a progress bar over batches and iterations.

    Returns
    -------
    scores : array of shape (n_targets, ) or (n_kernels, n_targets) (if split)
        Prediction score per target.
    """
    backend = get_backend()
    Ks, dual_weights, deltas, Y = backend.check_arrays(Ks, dual_weights,
                                                       deltas, Y)

    n_kernels, n_targets = deltas.shape
    if split:
        scores = backend.zeros_like(Y, shape=(n_kernels, n_targets))
    else:
        scores = backend.zeros_like(Y, shape=(n_targets))

    if n_targets_batch is None:
        n_targets_batch = n_targets
    for start in bar(list(range(0, n_targets, n_targets_batch)),
                     title='predict_and_score', use_it=progress_bar):
        batch = slice(start, start + n_targets_batch)
        predictions = predict_weighted_kernel_ridge(Ks, dual_weights[:, batch],
                                                    deltas[:, batch],
                                                    split=split)
        score_batch = score_func(Y[:, batch], predictions)

        if split:
            scores[:, batch] = score_batch
        else:
            scores[batch] = score_batch

    return scores


def primal_weights_kernel_ridge(dual_weights, X_fit):
    """Compute the primal weights for kernel ridge regression.

    Parameters
    ----------
    dual_weights : array of shape (n_samples_fit, n_targets)
        Dual coefficient of the kernel ridge regression.
    X_fit : array of shape (n_samples_fit, n_features)
        Training features.

    Returns
    -------
    primal_weights : array of shape (n_features, n_targets)
        Primal coefficients of the equivalent ridge regression. The
        coefficients are computed on CPU memory, since they can be large.
    """
    backend = get_backend()
    X_fit = backend.to_cpu(X_fit)
    dual_weights = backend.to_cpu(dual_weights)

    return X_fit.T @ dual_weights


def primal_weights_weighted_kernel_ridge(dual_weights, deltas, Xs_fit):
    """Compute the primal weights for weighted kernel ridge regression.

    Parameters
    ----------
    dual_weights : array of shape (n_samples_fit, n_targets)
        Dual coefficient of the kernel ridge regression.
    deltas : array of shape (n_kernels, n_targets)
        Log of kernel weights.
    Xs_fit : list of arrays of shape (n_samples_fit, n_features)
        Training features. The list should have `n_kernels` elements.

    Returns
    -------
    primal_weights : list of arrays of shape (n_features, n_targets)
        Primal coefficients of the equivalent ridge regression. The
        coefficients are computed on CPU memory, since they can be large.
    """
    backend = get_backend()
    dual_weights = backend.to_cpu(dual_weights)

    primal_weights = []
    for X_fit, deltas_i in zip(Xs_fit, deltas):
        X_fit = backend.to_cpu(X_fit)
        exp_deltas_i = backend.to_cpu(backend.exp(deltas_i))
        primal_weights_i = X_fit.T @ dual_weights * exp_deltas_i[None]
        primal_weights.append(primal_weights_i)

    return primal_weights



import pytest

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS

from himalaya.utils import compute_lipschitz_constants


@pytest.mark.parametrize('kernelize', ["XXT", "XTX", "X"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_compute_lipschitz_constants(backend, kernelize):
    backend = set_backend(backend)

    Xs = backend.randn(3, 5, 6)
    if kernelize == "X":
        XTs = backend.transpose(Xs, (0, 2, 1))
        Xs = backend.matmul(XTs, Xs)

    L = compute_lipschitz_constants(Xs)
    assert L.ndim == 1
    assert L.shape[0] == Xs.shape[0]


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_compute_lipschitz_constants_error(backend):
    backend = set_backend(backend)

    Xs = backend.randn(3, 5, 6)
    with pytest.raises(ValueError):
        compute_lipschitz_constants(Xs, "wrong")



import re
from pathlib import Path
from setuptools import find_packages, setup

# get version from himalaya/__init__.py
__version__ = 0.0
with open('himalaya/__init__.py') as f:
    infos = f.readlines()
for line in infos:
    if "__version__" in line:
        match = re.search(r"__version__ = '([^']*)'", line)
        __version__ = match.groups()[0]

# read the contents of the README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

requirements = [
    "numpy",
    "scikit-learn",
    # "cupy",  # optional backend
    # "torch",  # optional backend
    # "matplotlib",  # for visualization only
    # "pytest",  # for testing only
]

if __name__ == "__main__":
    setup(
        name='himalaya',
        maintainer="Tom Dupre la Tour",
        maintainer_email="tomdlt@berkeley.edu",
        description="Multiple-target machine learning",
        license='BSD (3-clause)',
        version=__version__,
        packages=find_packages(),
        url="https://github.com/gallantlab/himalaya",
        install_requires=requirements,
        long_description=long_description,
        long_description_content_type='text/x-rst'
    )



import types
import importlib
import warnings

#: List of all available backends.
ALL_BACKENDS = [
    "numpy",
    "cupy",
    "torch",
    "torch_cuda",
]

CURRENT_BACKEND = "numpy"


def set_backend(backend, on_error="raise"):
    """Set the backend using a global variable, and return the backend module.

    Parameters
    ----------
    backend : str or module
        Name or module of the backend.
    on_error : str in {"raise", "warn"}
        Define what is done if the backend fails to be loaded.
        If "warn", this function only warns, and keeps the previous backend.
        If "raise", this function raises on errors.

    Returns
    -------
    module : python module
        Module of the backend.
    """
    global CURRENT_BACKEND

    try:
        if isinstance(backend, types.ModuleType):  # get name from module
            backend = backend.name

        if backend not in ALL_BACKENDS:
            raise ValueError("Unknown backend=%r" % (backend, ))

        module = importlib.import_module(__package__ + "." + backend)
        CURRENT_BACKEND = backend
    except Exception as error:
        if on_error == "raise":
            raise error
        elif on_error == "warn":
            warnings.warn(f"Setting backend to {backend} failed: {str(error)}."
                          f"Falling back to {CURRENT_BACKEND} backend.")
            module = get_backend()
        else:
            raise ValueError('Unknown value on_error=%r' % (on_error, ))

    return module


def get_backend():
    """Get the current backend module.

    Returns
    -------
    module : python module
        Module of the backend.
    """
    module = importlib.import_module(__package__ + "." + CURRENT_BACKEND)
    return module


def _dtype_to_str(dtype):
    """Cast dtype to string, such as "float32", or "float64"."""
    if isinstance(dtype, str):
        return dtype
    elif hasattr(dtype, "name"):  # works for numpy and cupy
        return dtype.name
    elif "torch." in str(dtype):  # works for torch
        return str(dtype)[6:]
    elif dtype is None:
        return None
    else:
        raise NotImplementedError()



"""
Kernel ridge with cross-validation
==================================

This example demonstrates how to solve kernel ridge regression with a
cross-validation of the regularization parameter, using himalaya's estimator
``KernelRidgeCV``.
"""

###############################################################################
# Create a random dataset
# -----------------------
import numpy as np
np.random.seed(0)
n_samples, n_features, n_targets = 10, 5, 4
X = np.random.randn(n_samples, n_features)
Y = np.random.randn(n_samples, n_targets)

###############################################################################
# Limit of GridSearchCV
# ---------------------
# In scikit-learn, one can use ``GridSearchCV`` to optimize hyperparameters
# over cross-validation.

import sklearn.model_selection
import sklearn.kernel_ridge

estimator = sklearn.kernel_ridge.KernelRidge(kernel="linear")
gscv = sklearn.model_selection.GridSearchCV(
    estimator=estimator,
    param_grid=dict(alpha=np.logspace(-2, 2, 5)),
)
gscv.fit(X, Y)

###############################################################################
# However, since ``GridSearchCV`` optimizes the average score over all targets,
# it returns a single value for alpha.
gscv.best_params_

###############################################################################
# KernelRidgeCV
# -------------
# To optimize each target independently, himalaya implements ``KernelRidgeCV``,
# which supports any cross-validation scheme compatible with scikit-learn.
import himalaya.kernel_ridge

model = himalaya.kernel_ridge.KernelRidgeCV(kernel="linear",
                                            alphas=np.logspace(-2, 2, 5))
model.fit(X, Y)

###############################################################################
# KernelRidgeCV returns a separate best alpha per target.
model.best_alphas_



import pytest

import sklearn.linear_model

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.lasso import solve_sparse_group_lasso
from himalaya.lasso import solve_sparse_group_lasso_cv


def _create_dataset(backend):
    n_samples, n_features, n_targets = 10, 5, 3

    X = backend.asarray(backend.randn(n_samples, n_features), backend.float64)
    Y = backend.asarray(backend.randn(n_samples, n_targets), backend.float64)

    return X, Y


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_vs_ols(backend):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    coef = solve_sparse_group_lasso(X, Y, groups=None, l21_reg=0.0, l1_reg=0.0,
                                    max_iter=1000, tol=1e-8,
                                    progress_bar=False)

    ols = sklearn.linear_model.LinearRegression(fit_intercept=False).fit(
        backend.to_numpy(X), backend.to_numpy(Y))
    assert_array_almost_equal(coef, ols.coef_.T, decimal=4)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_decreasing(backend):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    coef, losses = solve_sparse_group_lasso(X, Y, max_iter=500, tol=1e-8,
                                            progress_bar=False, debug=True,
                                            momentum=False)

    assert backend.all(losses[1:] - losses[:-1] < 1e-14)


@pytest.mark.parametrize('n_targets_batch', [None, 2])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_vs_lasso(backend, n_targets_batch):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    for l1_reg in backend.logspace(-5, 5, 5):

        coef = solve_sparse_group_lasso(X, Y, groups=None, l21_reg=0.0,
                                        l1_reg=l1_reg, max_iter=1000, tol=1e-8,
                                        progress_bar=False, debug=False,
                                        momentum=False,
                                        n_targets_batch=n_targets_batch)

        ols = sklearn.linear_model.Lasso(fit_intercept=False,
                                         alpha=backend.to_numpy(l1_reg),
                                         max_iter=1000,
                                         tol=1e-8).fit(backend.to_numpy(X),
                                                       backend.to_numpy(Y))
        assert_array_almost_equal(coef, ols.coef_.T, decimal=5)


@pytest.mark.parametrize('n_targets_batch', [None, 2])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_regularization_per_target(backend, n_targets_batch):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    n_targets = Y.shape[1]
    l21_reg = backend.rand(n_targets)
    l1_reg = backend.rand(n_targets)

    coef = solve_sparse_group_lasso(X, Y, groups=None, l21_reg=l21_reg,
                                    l1_reg=l1_reg, max_iter=1000, tol=1e-8,
                                    progress_bar=False, debug=False,
                                    momentum=False,
                                    n_targets_batch=n_targets_batch)

    for tt in range(n_targets):

        coef_tt = solve_sparse_group_lasso(X, Y[:, tt:tt + 1], groups=None,
                                           l21_reg=l21_reg[tt],
                                           l1_reg=l1_reg[tt], max_iter=1000,
                                           tol=1e-8, progress_bar=False,
                                           debug=False, momentum=False,
                                           n_targets_batch=n_targets_batch)
        assert_array_almost_equal(coef[:, tt:tt + 1], coef_tt)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_cv(backend):
    backend = set_backend(backend)
    X, Y = _create_dataset(backend)

    n_targets = Y.shape[1]
    l21_regs = backend.rand(2) / 10
    l1_regs = backend.rand(3) / 10

    coef, best_l21_reg, best_l1_reg, all_cv_scores = \
        solve_sparse_group_lasso_cv(
            X, Y, cv=2, groups=None, l21_regs=l21_regs, l1_regs=l1_regs,
            progress_bar=False, tol=1e-2, max_iter=100)

    assert best_l1_reg.shape == (n_targets, )
    assert best_l21_reg.shape == (n_targets, )
    assert coef.shape == (X.shape[1], n_targets)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_group_lasso_group_sparsity(backend):
    backend = set_backend(backend)

    # Set dataset parameters
    import numpy as np
    group_sizes = [np.random.randint(10, 20) for i in range(5)]
    active_groups = [np.random.randint(2) for _ in group_sizes]
    active_groups[1] = 1  # make sure we have at least one active...
    active_groups[0] = 0  # ...and one inactive group
    groups = np.concatenate([size * [i] for i, size in enumerate(group_sizes)])
    n_features = sum(group_sizes)
    n_samples = 10000
    noise_std = 10

    # Generate data matrix
    X = backend.randn(n_samples, n_features)
    w = backend.concatenate([
        backend.randn(group_size) * is_active
        for group_size, is_active in zip(group_sizes, active_groups)
    ]).reshape(-1, 1)
    y = X @ w
    y = y + backend.randn(*y.shape) * noise_std

    # Generate estimator and train it
    coef = solve_sparse_group_lasso(X=X, Y=y, groups=groups, l21_reg=0.6,
                                    l1_reg=0, max_iter=100, tol=1e-4,
                                    progress_bar=False)

    # check the group sparsity of the result
    for group, active in enumerate(active_groups):
        if active:
            assert backend.all(coef[groups == group] != 0)
        else:
            assert backend.all(coef[groups == group] == 0)



import numpy as np
from sklearn.decomposition import PCA

import matplotlib.pyplot as plt
from matplotlib.patches import Polygon

from himalaya.kernel_ridge import generate_dirichlet_samples
kernel_weights = generate_dirichlet_samples(10000, n_kernels=3,
                                            concentration=[1.], random_state=0)
pca = PCA(2).fit(kernel_weights)

darkgreen = "#446455"
white = "white"


def plot_simplex(bias=(0, 0), ax=None):
    corners = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    corners = pca.transform(corners).T

    if ax is None:
        plt.figure(figsize=(2, 2))
        ax = plt.gca()

    # Faces
    ax.add_patch(
        Polygon(corners[:2].T + bias, closed=True, edgecolor=None, fill=True,
                facecolor=white, alpha=0.6))
    # Edges
    ax.add_patch(
        Polygon(corners[:2].T + bias, closed=True, edgecolor=darkgreen,
                fill=False, alpha=1, linewidth=2))

    ax.axis('equal')
    ax.axis('off')
    return ax


fig, ax = plt.subplots(figsize=(2, 2))

bias = [0.4, 0.04]
for factor in np.linspace(2, 0, 3):
    bias_ = np.array(bias) * factor
    plot_simplex(bias_, ax=ax)

ax.text(-0.4, -1, "himalaya", fontsize=20, color=darkgreen)
fig.savefig("logo.svg", bbox_inches='tight', pad_inches=0)



import pytest
import sklearn.metrics

from himalaya.backend import set_backend
from himalaya.backend import ALL_BACKENDS
from himalaya.utils import assert_array_almost_equal

from himalaya.scoring import r2_score
from himalaya.scoring import l2_neg_loss
from himalaya.scoring import correlation_score
from himalaya.scoring import r2_score_split


def _create_data(backend, dtype_str):
    n_alphas, n_samples, n_targets = 10, 20, 30
    y_pred = backend.asarray(backend.randn(n_alphas, n_samples, n_targets),
                             dtype=getattr(backend, dtype_str))
    y_true = backend.asarray(backend.randn(n_samples, n_targets),
                             dtype=getattr(backend, dtype_str))
    y_true -= y_true.mean(0)
    return y_pred, y_true


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_r2_score(backend, dtype_str):
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    # multi predictions
    s_1 = backend.stack([
        backend.asarray(
            sklearn.metrics.r2_score(backend.to_numpy(y_true),
                                     backend.to_numpy(pp),
                                     multioutput='raw_values'),
            dtype=dtype_str) for pp in y_pred
    ])
    s_2 = r2_score(y_true, y_pred)
    assert_array_almost_equal(s_1, s_2, decimal=5)

    # single prediction
    s_1 = backend.asarray(
        sklearn.metrics.r2_score(backend.to_numpy(y_true),
                                 backend.to_numpy(y_pred[0]),
                                 multioutput='raw_values'), dtype=dtype_str)
    s_2 = r2_score(y_true, y_pred[0])
    assert_array_almost_equal(s_1, s_2, decimal=5)


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_r2_score_split(backend, dtype_str):
    # r2_score_split(include_correlation=False) is equivalent to r2_score
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    # multi predictions
    s_1 = backend.stack([
        backend.asarray(
            sklearn.metrics.r2_score(backend.to_numpy(y_true),
                                     backend.to_numpy(pp),
                                     multioutput='raw_values'),
            dtype=dtype_str) for pp in y_pred
    ])
    s_2 = r2_score_split(y_true, y_pred, False)
    assert_array_almost_equal(s_1, s_2, decimal=5)

    # single prediction
    s_1 = backend.asarray(
        sklearn.metrics.r2_score(backend.to_numpy(y_true),
                                 backend.to_numpy(y_pred[0]),
                                 multioutput='raw_values'), dtype=dtype_str)
    s_2 = r2_score_split(y_true, y_pred[0], False)
    assert_array_almost_equal(s_1, s_2, decimal=5)


@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_r2_score_split_warn_non_zero_mean(backend):
    # r2_score_split requires y_true to be zero mean
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, "float32")

    with pytest.warns(UserWarning):
        _ = r2_score_split(y_true + 1, y_pred, False)
    with pytest.warns(UserWarning):
        _ = r2_score_split(y_true + 1, y_pred[0], False)


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_r2_score_split_include_correlation(backend, dtype_str):
    # r2_score_split(include_correlation=True) gives results that sum to the
    # r2 score of the sum.
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    s_1 = r2_score_split(y_true, y_pred.sum(0), True)
    s_2 = r2_score_split(y_true, y_pred, True)
    assert_array_almost_equal(s_1, s_2.sum(0), decimal=5)


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_l2_neg_loss(backend, dtype_str):
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    # multi predictions
    s_1 = -backend.stack([
        backend.asarray(
            sklearn.metrics.mean_squared_error(
                backend.to_numpy(y_true), backend.to_numpy(pp),
                multioutput='raw_values'), dtype=dtype_str) for pp in y_pred
    ])
    s_2 = l2_neg_loss(y_true, y_pred)
    assert_array_almost_equal(s_1, s_2, decimal=5)

    # single prediction
    s_1 = -backend.asarray(
        sklearn.metrics.mean_squared_error(
            backend.to_numpy(y_true), backend.to_numpy(
                y_pred[0]), multioutput='raw_values'), dtype=dtype_str)
    s_2 = l2_neg_loss(y_true, y_pred[0])
    assert_array_almost_equal(s_1, s_2, decimal=5)


def _correlation(backend, y_true, y_pred, dtype):
    y_true = backend.asarray(y_true, dtype="float64")
    y_pred = backend.asarray(y_pred, dtype="float64")

    y_pred -= y_pred.mean(0)
    y_pred -= y_pred.mean(0)

    y_true /= backend.std_float64(y_true, 0)
    y_pred /= backend.std_float64(y_pred, 0)

    return backend.asarray((y_true * y_pred).mean(0), dtype=dtype)


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_correlation_score(backend, dtype_str):
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    # multi predictions
    s_1 = backend.stack(
        [_correlation(backend, y_true, pp, dtype=dtype_str) for pp in y_pred])
    s_2 = correlation_score(y_true, y_pred)
    assert_array_almost_equal(s_1, s_2, decimal=5)

    # single prediction
    s_1 = _correlation(backend, y_true, y_pred[0], dtype=dtype_str)
    s_2 = correlation_score(y_true, y_pred[0])
    assert_array_almost_equal(s_1, s_2, decimal=5)


@pytest.mark.parametrize('dtype_str', ["float32", "float64"])
@pytest.mark.parametrize('scoring', [
    r2_score,
    l2_neg_loss,
    correlation_score,
    r2_score_split,
])
@pytest.mark.parametrize('backend', ALL_BACKENDS)
def test_infinite_and_nans(backend, dtype_str, scoring):
    backend = set_backend(backend)
    y_pred, y_true = _create_data(backend, dtype_str)

    # nan as zero
    y_pred[0][10, 10] = backend.nan
    with pytest.warns(UserWarning) as record:
        s_2 = scoring(y_true, y_pred[0])
    assert not backend.any(backend.isnan(s_2))
    assert len(record) == 1
    assert record[0].message.args[0] == "nan in y_pred."

    # inf as zero
    y_pred[0][10, 10] = backend.inf
    with pytest.warns(UserWarning) as record:
        s_2 = scoring(y_true, y_pred[0])
    assert not backend.any(backend.isnan(s_2))
    assert len(record) == 1
    assert record[0].message.args[0] == "inf in y_pred."



"""
Multiple kernel ridge solvers
=============================
This example demonstrates the different strategies to solve the multiple kernel
ridge regression: the *random search*, and the *hyper-gradient descent*.

The *random-search* strategy samples some kernel weights vectors from a Dirichlet
distribution, then for each vector, it fits a ``KernelRidgeCV`` model and
computes a cross-validation score for all targets. Then it selects for each
target the kernel weight vector leading to the highest cross-validation score
(e.g. the highest R\ :sup:`2` value).
Extensively sampling the kernel weights space is exponentially expensive with
the number of kernels, therefore this method is computationally expensive for a
large number of kernels. However, since it reuses most of the computations for
all targets, it scales very well with the number of targets.

The *hyper-gradient descent* strategy takes a different route. It starts with
an initial kernel weights vector per target, and updates it iteratively
following the hyperparameter gradient, computed over cross-validation. As it
computes a hyper-gradient descent for each target, it is more expensive
computationally for large number of targets. However, the hyper-gradient
descent scales very well with the number of kernels.
"""
import numpy as np
import matplotlib.pyplot as plt

from himalaya.backend import set_backend
from himalaya.kernel_ridge import generate_dirichlet_samples

from himalaya.kernel_ridge import KernelRidgeCV
from himalaya.kernel_ridge import MultipleKernelRidgeCV
from himalaya.kernel_ridge import Kernelizer
from himalaya.kernel_ridge import ColumnKernelizer

from sklearn.pipeline import make_pipeline
from sklearn import set_config
set_config(display='diagram')

# sphinx_gallery_thumbnail_number = 4
###############################################################################
# In this example, we use the ``torch`` backend, and fit the model on GPU.

backend = set_backend("torch_cuda")

###############################################################################
# Generate a random dataset
# -------------------------
# We start by generating some arbitrary scalings per kernel and targets, using
# samples on a Dirichlet distribution.

n_kernels = 3
n_targets = 50
n_clusters = 2

###############################################################################
# To create some clusters of weights, we take a few kernel weights samples..
kernel_weights_true = generate_dirichlet_samples(n_clusters, n_kernels,
                                                 concentration=[.3],
                                                 random_state=105)

###############################################################################
# .. then, we duplicate them, and add some noise, to get clusters.
noise = 0.05
kernel_weights_true = backend.to_numpy(kernel_weights_true)
kernel_weights_true = np.tile(kernel_weights_true,
                              (n_targets // n_clusters, 1))
kernel_weights_true += np.random.randn(n_targets, n_kernels) * noise

# We finish with a projection on the simplex, making kernel weights sum to one.
kernel_weights_true[kernel_weights_true < 0] = 0.
kernel_weights_true /= np.sum(kernel_weights_true, 1)[:, None]

###############################################################################
# Then, we generate a random dataset, using the arbitrary scalings.
#
# - Xs_train : list of arrays of shape (n_samples_train, n_features)
# - Xs_test : list of arrays of shape (n_samples_test, n_features)
# - Y_train : array of shape (n_samples_train, n_targets)
# - Y_test : array of shape (n_repeat, n_samples_test, n_targets)

n_samples_train = 1000
n_samples_test = 300
n_features_list = np.full(n_kernels, fill_value=1000)

Xs_train, Xs_test = [], []
Y_train, Y_test = None, None
for ii in range(n_kernels):
    n_features = n_features_list[ii]

    X_train = backend.randn(n_samples_train, n_features)
    X_test = backend.randn(n_samples_test, n_features)
    Xs_train.append(X_train)
    Xs_test.append(X_test)

    weights = backend.randn(n_features, n_targets) / n_features
    weights *= backend.asarray_like(kernel_weights_true[:, ii],
                                    ref=weights) ** 0.5

    if ii == 0:
        Y_train = X_train @ weights
        Y_test = X_test @ weights
    else:
        Y_train += X_train @ weights
        Y_test += X_test @ weights

std = Y_train.std(0)[None]
Y_train /= std
Y_test /= std

noise = 0.1
Y_train += backend.randn(n_samples_train, n_targets) * noise
Y_test += backend.randn(n_samples_test, n_targets) * noise

###############################################################################
# Define a ``ColumnKernelizer``
# -----------------------------
# We define a column kernelizer, which we will use to precompute the kernels in
# a pipeline.

feature_names = ["space %d" % ii for ii in range(n_kernels)]

###############################################################################
# Concatenate the feature spaces
X_train = backend.asarray(backend.concatenate(Xs_train, 1), dtype="float32")
X_test = backend.asarray(backend.concatenate(Xs_test, 1), dtype="float32")

###############################################################################
# Find the start and end of each feature space X in Xs
start_and_end = np.concatenate([[0], np.cumsum(n_features_list)])
slices = [
    slice(start, end)
    for start, end in zip(start_and_end[:-1], start_and_end[1:])
]

kernelizers = [(name, Kernelizer(), slice_)
               for name, slice_ in zip(feature_names, slices)]
column_kernelizer = ColumnKernelizer(kernelizers)

###############################################################################
# Define the models
# -----------------
# We define the first model, using the random search solver.

# (We pregenerate the Dirichlet random samples, to latter plot them.)
kernel_weights_sampled = generate_dirichlet_samples(n_samples=20,
                                                    n_kernels=n_kernels,
                                                    concentration=[1.],
                                                    random_state=0)

alphas = np.logspace(-10, 10, 41)
solver_params = dict(n_iter=kernel_weights_sampled, alphas=alphas,
                     n_targets_batch=200, n_alphas_batch=20,
                     n_targets_batch_refit=200, jitter_alphas=True)

model_1 = MultipleKernelRidgeCV(kernels="precomputed", solver="random_search",
                                solver_params=solver_params)

###############################################################################
# We define the second model, using the hyper_gradient solver.

solver_params = dict(max_iter=80, n_targets_batch=200, tol=1e-3,
                     initial_deltas="ridgecv", max_iter_inner_hyper=1,
                     hyper_gradient_method="direct")

model_2 = MultipleKernelRidgeCV(kernels="precomputed", solver="hyper_gradient",
                                solver_params=solver_params)

###############################################################################
# We fit the two models on the train data.

pipe_1 = make_pipeline(column_kernelizer, model_1)
pipe_1.fit(X_train, Y_train)

###############################################################################
pipe_2 = make_pipeline(column_kernelizer, model_2)
pipe_2.fit(X_train, Y_train)

###############################################################################
# Plot the convergence curves
# ---------------------------
# First convergence curve.
#
# For the random search, ``cv_scores`` gives the scores for each sampled kernel
# weights vector. The convergence curve is thus the current maximum for each target.
cv_scores = backend.to_numpy(pipe_1[1].cv_scores_)
current_max = np.maximum.accumulate(cv_scores, axis=0)
mean_current_max = np.mean(current_max, axis=1)

x_array = np.arange(1, len(mean_current_max) + 1)
plt.plot(x_array, mean_current_max, '-o')
plt.grid("on")
plt.xlabel("Number of kernel weights sampled")
plt.ylabel("L2 negative loss (higher is better)")
plt.title("Convergence curve, averaged over targets")
plt.show()

###############################################################################
# Plot the second convergence curve.

cv_scores = backend.to_numpy(pipe_2[1].cv_scores_)
mean_cv_scores = np.mean(cv_scores, axis=1)

x_array = np.arange(1, len(mean_cv_scores) + 1)
plt.plot(x_array, mean_cv_scores, '-o')
plt.grid("on")
plt.xlabel("Number of gradient iterations")
plt.ylabel("L2 negative loss (higher is better)")
plt.title("Convergence curve, averaged over targets")
plt.show()

###############################################################################
# Compare with a ``KernelRidgeCV``
# --------------------------------
# Compare to a baseline ``KernelRidgeCV`` model with all the concatenated features.
# Comparison is performed using the prediction scores on the test set.

# Fit the baseline model ``KernelRidgeCV``
baseline = KernelRidgeCV(kernel="linear", alphas=alphas)
baseline.fit(X_train, Y_train)

###############################################################################
# Compute scores of all models
scores_1 = pipe_1.score(X_test, Y_test)
scores_1 = backend.to_numpy(scores_1)

scores_2 = pipe_2.score(X_test, Y_test)
scores_2 = backend.to_numpy(scores_2)

scores_baseline = baseline.score(X_test, Y_test)
scores_baseline = backend.to_numpy(scores_baseline)

###############################################################################
# Plot histograms
bins = np.linspace(
    np.min([scores_baseline.min(),
            scores_1.min(),
            scores_2.min()]),
    np.max([scores_baseline.max(),
            scores_1.max(),
            scores_2.max()]), 25)
plt.hist(scores_1, bins, alpha=0.5,
         label="MultipleKernelRidgeCV(solver='random_search')")
plt.hist(scores_2, bins, alpha=0.5,
         label="MultipleKernelRidgeCV(solver='hyper_gradient')")
plt.hist(scores_baseline, bins, alpha=0.5, label="KernelRidgeCV")
plt.xlabel(r"$R^2$ generalization score")
plt.title("Histogram over targets")
plt.legend()
plt.show()

###############################################################################
# Generate trajectories
# ---------------------
# Refit the second model with different number of iterations, just to plot the
# trajectories.

all_kernel_weights_2 = [
    np.full((n_targets, n_kernels), fill_value=1. / n_kernels),
]
for max_iter in np.unique(np.int_(np.logspace(0, np.log10(80), 3))):
    # change max_iter and refit from scratch
    pipe_2[1].solver_params['max_iter'] = max_iter
    pipe_2.fit(X_train, Y_train)

    kernel_weights_2 = np.exp(backend.to_numpy(pipe_2[1].deltas_.T))
    kernel_weights_2 /= kernel_weights_2.sum(1)[:, None]
    all_kernel_weights_2.append(kernel_weights_2)

###############################################################################
# Get the normalized kernel weights for the first model
kernel_weights_1 = np.exp(backend.to_numpy(pipe_1[1].deltas_.T))
kernel_weights_1 /= kernel_weights_1.sum(1)[:, None]

###############################################################################
# Plot on the simplex
# -------------------
# Finally, we visualize the obtained kernel weights vector, projected on the simplex.
# The simplex is the space of positive weights that sum to one, and it has a
# triangular shape in dimension 3.
# We plot on three different panels:
#
# - the kernel weights used in the simulated data
# - the kernel weights sampled during random search, and the best ones
# - the kernel weights trajectories obtained during hyper-gradient descent


def _create_simplex_projection_and_edges(ax):
    """Create a projection on the 3D simplex, and plot edges."""
    n_kernels = 3

    if ax is None:
        ax = plt.gca()

    # create a projection in 2D
    from sklearn.decomposition import PCA
    kernel_weights = generate_dirichlet_samples(10000, n_kernels,
                                                concentration=[1.],
                                                random_state=0)
    pca = PCA(2).fit(backend.to_numpy(kernel_weights))

    # add simplex edges
    edges = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 0, 0]])
    edges = pca.transform(edges).T

    # add tripod at origin
    tripod_length = 0.15
    tripod = np.array([[0, 0, 0], [tripod_length, 0, 0], [0, 0, 0],
                       [0, tripod_length, 0], [0, 0, 0], [0, 0,
                                                          tripod_length]])
    tripod = pca.transform(tripod).T

    # add point legend
    points = np.array([[1, 0, 0], [0, 1, 0],
                       [0, 0, 1]])
    labels = points.copy()
    points = pca.transform(points * 1.15).T
    for (xx, yy), label in zip(points.T, labels):
        ax.text(xx, yy, str(label), horizontalalignment='center',
                verticalalignment='center')

    if ax is None:
        plt.figure(figsize=(8, 8))
        ax = plt.gca()
    ax.plot(edges[0], edges[1], c='gray')
    ax.plot(tripod[0], tripod[1], c='gray')
    ax.axis('equal')
    ax.axis('off')
    return ax, pca


def plot_simplex(X, ax=None, **kwargs):
    """Plot a set of points in the 3D simplex."""
    ax, pca = _create_simplex_projection_and_edges(ax=ax)

    Xt = pca.transform(X).T
    ax.scatter(Xt[0], Xt[1], **kwargs)
    ax.legend()
    return ax


def plot_simplex_trajectory(Xs, ax=None):
    """Plot a series of trajectory in the 3D simplex."""
    ax, pca = _create_simplex_projection_and_edges(ax=ax)

    trajectories = []
    for Xi in Xs:
        Xt = pca.transform(Xi).T
        trajectories.append(Xt)
    trajectories = np.array(trajectories)

    for trajectory in trajectories.T:
        ax.plot(trajectory[0], trajectory[1], linewidth=1, color="C0",
                zorder=1)
        ax.scatter(trajectory[0, -1], trajectory[1, -1], color="C1", zorder=2)

    return ax


###############################################################################
fig, axs = plt.subplots(1, 3, figsize=(12, 4))

# First panel
ax = axs[0]
ax.set_title("(a) Ground truth", y=0)
plot_simplex(kernel_weights_true, ax=ax, color='C2',
             label="true weights")

# Second panel
ax = axs[1]
ax.set_title("(b) Random search", y=0)
plot_simplex(backend.to_numpy(kernel_weights_sampled), ax=ax, marker='+',
             label="random candidates", zorder=10)
plot_simplex(kernel_weights_1, ax=axs[1], label="selected candidates")

# Third panel
ax = axs[2]
ax.set_title("(c) Gradient descent", y=0)
plot_simplex_trajectory(all_kernel_weights_2, ax=ax)
ax.legend([ax.lines[2], ax.collections[0]],
          ['gradient trajectory', 'final point'])

plt.tight_layout()
# fig.savefig('simulation.pdf', dpi=150, bbox_inches='tight', pad_inches=0)
plt.show()



"""
Kernel ridge regression
=======================

This example demonstrates how to solve kernel ridge regression, using
himalaya's estimator ``KernelRidge`` compatible with scikit-learn's API.
"""

###############################################################################
# Create a random dataset
# -----------------------
import numpy as np
n_samples, n_features, n_targets = 10, 5, 4
X = np.random.randn(n_samples, n_features)
Y = np.random.randn(n_samples, n_targets)

###############################################################################
# Scikit-learn API
# ----------------
# Himalaya implements a ``KernelRidge`` estimator, similar to the corresponding
# scikit-learn estimator, with similar parameters and methods.
import sklearn.kernel_ridge
import himalaya.kernel_ridge

# Fit a scikit-learn model
model_skl = sklearn.kernel_ridge.KernelRidge(kernel="linear", alpha=0.1)
model_skl.fit(X, Y)

# Fit a himalaya model
model_him = himalaya.kernel_ridge.KernelRidge(kernel="linear", alpha=0.1)
model_him.fit(X, Y)

Y_pred_skl = model_skl.predict(X)
Y_pred_him = model_him.predict(X)

# The predictions are virtually identical.
print(np.max(np.abs(Y_pred_skl - Y_pred_him)))

###############################################################################
# Small API difference
# --------------------
# Since himalaya focuses on fitting multiple targets, the ``score`` method
# returns the score on each target separately, while scikit-learn returns the
# average score over targets.

print(model_skl.score(X, Y))
print(model_him.score(X, Y))
print(model_him.score(X, Y).mean())
